import importlib
import os
from functools import wraps
from pathlib import Path
from typing import Optional, List, Type, Callable

from flask import Flask, request
from flask_jwt_extended import jwt_required
from flask_limiter import Limiter
from flask_limiter.util import get_remote_address
from marshmallow import Schema
from sqlalchemy.orm import DeclarativeBase

from flarchitect.authentication.jwt import get_user_from_token
from flarchitect.authentication.user import set_current_user
from flarchitect.exceptions import CustomHTTPException
from flarchitect.logging import logger
from flarchitect.utils.general import (
    AttributeInitializerMixin,
    check_rate_services,
    validate_flask_limiter_rate_limit_string,
)
from flarchitect.utils.config_helpers import get_config_or_model_meta
from flarchitect.utils.decorators import handle_many, handle_one
from flarchitect.core.routes import find_rule_by_function, RouteCreator
from flarchitect.specs.generator import CustomSpec

FLASK_APP_NAME = "flarchitect"


class Architect(AttributeInitializerMixin):
    app: Flask
    api_spec: Optional[CustomSpec] = None
    api: Optional["RouteCreator"] = None
    base_dir: str = os.path.dirname(os.path.abspath(__file__))
    route_spec: List = []
    limiter: Limiter

    def __init__(self, app: Optional[Flask] = None, *args, **kwargs):
        """
        Initializes the Architect object.

        Args:
            app (Flask): The flask app.
            *args (list): List of arguments.
            **kwargs (dict): Dictionary of keyword arguments.
        """
        if app is not None:
            self.init_app(app, *args, **kwargs)
            logger.verbosity_level = self.get_config("API_VERBOSITY_LEVEL", 0)

    def init_app(self, app: Flask, *args, **kwargs):
        """
        Initializes the Architect object.

        Args:
            app (Flask): The flask app.
            *args (list): List of arguments.
            **kwargs (dict): Dictionary of keyword arguments.
        """
        super().__init__(app=app, *args, **kwargs)
        self._register_app(app)
        logger.verbosity_level = self.get_config("API_VERBOSITY_LEVEL", 0)
        self.api_spec = None

        if self.get_config("FULL_AUTO", True):
            self.init_api(app=app, **kwargs)
        if get_config_or_model_meta("API_CREATE_DOCS", default=True):
            self.init_apispec(app=app, **kwargs)

        logger.log(2, "Creating rate limiter")
        storage_uri = check_rate_services()
        self.app.config["RATELIMIT_HEADERS_ENABLED"] = True
        self.app.config["RATELIMIT_SWALLOW_ERRORS"] = True
        self.app.config["RATELIMIT_IN_MEMORY_FALLBACK_ENABLED"] = True
        self.limiter = Limiter(
            app=app,
            key_func=get_remote_address,
            storage_uri=storage_uri if storage_uri else None,
        )

        @app.teardown_request
        def clear_current_user(exception=None):
            set_current_user(None)

    def _register_app(self, app: Flask):
        """
        Registers the app with the extension, and saves it to self.

        Args:
            app (Flask): The flask app.
        """
        if FLASK_APP_NAME not in app.extensions:
            app.extensions[FLASK_APP_NAME] = self
        self.app = app

    def init_apispec(self, app: Flask, **kwargs):
        """
        Initializes the api spec object.

        Args:
            app (Flask): The flask app.
            **kwargs (dict): Dictionary of keyword arguments.
        """
        self.api_spec = CustomSpec(app=app, architect=self, **kwargs)

    def init_api(self, **kwargs):
        """
        Initializes the api object, which handles flask route creation for models.

        Args:
            **kwargs (dict): Dictionary of keyword arguments.
        """
        self.api = RouteCreator(architect=self, **kwargs)

    def to_api_spec(self):
        """
        Returns the api spec object.

        Returns:
            APISpec: The api spec json object.
        """
        if self.api_spec:
            return self.api_spec.to_dict()

    def get_config(self, key, default: Optional = None):
        """
        Gets a config value from the app config.

        Args:
            key (str): The key of the config value.
            default (Optional): The default value to return if the key is not found.

        Returns:
            Any: The config value.
        """
        if self.app:
            return self.app.config.get(key, default)

    def schema_constructor(
        self,
        output_schema: Optional[Type[Schema]] = None,
        input_schema: Optional[Type[Schema]] = None,
        model: Optional[DeclarativeBase] = None,
        group_tag: Optional[str] = None,
        many: Optional[bool] = False,
        **kwargs,
    ) -> Callable:
        """
        Decorator to specify OpenAPI metadata for an endpoint, along with schema information for the input and output.

        Args:
            output_schema (Optional[Type[Schema]], optional): Output schema. Defaults to None.
            input_schema (Optional[Type[Schema]], optional): Input schema. Defaults to None.
            model (Optional[DeclarativeBase], optional): Database model. Defaults to None.
            group_tag (Optional[str], optional): Group name. Defaults to None.
            many (Optional[Bool], optional): Is many or one? Defaults to False
            kwargs (dict): Dictionary of keyword arguments.

        Returns:
            Callable: The decorated function.
        """

        def decorator(f: Callable) -> Callable:
            @wraps(f)
            def wrapped(*_args, **_kwargs):
                f_decorated = f

                auth_method = get_config_or_model_meta(
                    "API_AUTHENTICATE",
                    model=model,
                    output_schema=output_schema,
                    input_schema=input_schema,
                    method=request.method,
                    default=False,
                )

                def jwt_authentication(func):
                    @wraps(func)
                    def auth_wrapped(*args, **kwargs):
                        try:
                            auth = request.headers.get("Authorization")
                            if auth:
                                token = auth.split(" ")[1]
                                usr = get_user_from_token(token, secret_key=None)
                                if usr:
                                    set_current_user(usr)
                                    return func(*args, **kwargs)
                        except CustomHTTPException:
                            return CustomHTTPException.status_code

                    return auth_wrapped


                if auth_method:
                    if not isinstance(auth_method, list):
                        auth_method = [auth_method]

                decorated_f = None
                for method in auth_method:

                    if method == "jwt" and not decorated_f:
                        decorated_f = jwt_authentication(f_decorated)
                    if method == "basic" and not decorated_f:
                        decorated_f = None  # authentication(f_decorated)
                    if method == "api_key" and not decorated_f:
                        decorated_f = None  # authentication(f_decorated)
                    if method == "custom" and not decorated_f:
                        decorated_f = get_config_or_model_meta("API_CUSTOM_AUTH")(f_decorated)

                    if isinstance(decorated_f, int):
                        raise CustomHTTPException(status_code=f_decorated)
                    if callable(decorated_f):
                        f_decorated = decorated_f

                f_decorated = (
                    handle_many(output_schema, input_schema)(f_decorated)
                    if many
                    else handle_one(output_schema, input_schema)(f_decorated)
                )

                rl = get_config_or_model_meta(
                    "API_RATE_LIMIT",
                    model=model,
                    input_schema=input_schema,
                    output_schema=output_schema,
                    default=False,
                )
                if (
                    rl
                    and isinstance(rl, str)
                    and validate_flask_limiter_rate_limit_string(rl)
                ):
                    f_decorated = self.limiter.limit(rl)(f_decorated)
                elif rl:
                    rule = find_rule_by_function(self, f).rule
                    logger.error(
                        f"Rate limit definition not a string or not valid. Skipping for `{rule}` route."
                    )

                result = f_decorated(*_args, **_kwargs)
                return result

            route_info = {
                "function": wrapped,
                "output_schema": output_schema,
                "input_schema": input_schema,
                "model": model,
                "group_tag": group_tag,
                "tag": kwargs.get("tag"),
                "summary": kwargs.get("summary"),
                "error_responses": kwargs.get("error_responses"),
                "many_to_many_model": kwargs.get("many_to_many_model"),
                "multiple": many or kwargs.get("multiple"),
                "parent": kwargs.get("parent_model"),
            }

            self.set_route(route_info)
            return wrapped

        return decorator

    @classmethod
    def get_templates_path(cls, folder_name="html", max_levels=3):
        """
        Recursively searches for the folder_name in the source directory
        or its parent directories up to max_levels levels.

        Args:
            folder_name (str): The name of the folder to search for. Default is "html".
            max_levels (int): Maximum number of levels to search upwards. Default is 3.

        Returns:
            str: The path to the folder if found, else None.
        """
        # Find the source directory of the class/module
        spec = importlib.util.find_spec(cls.__module__)
        source_dir = Path(os.path.split(spec.origin)[0])

        # Traverse up to max_levels levels
        for level in range(max_levels):
            # Search for the folder in the current directory
            potential_path = source_dir / folder_name
            if potential_path.exists() and potential_path.is_dir():
                return str(potential_path)

            # Move to the parent directory for the next level search
            source_dir = source_dir.parent

        # Return None if the folder is not found
        return None

    def set_route(self, route: dict):
        """
        Adds a route to the route spec list, which is used to generate the api spec.

        Args:
            route (dict): The route object.
        """
        if not hasattr(route["function"], "_decorators"):
            route["function"]._decorators = []

        route["function"]._decorators.append(self.schema_constructor)
        self.route_spec.append(route)



from marshmallow import Schema, fields, pre_load, post_load
from sqlalchemy.testing.pickleable import User


class DeleteSuccessfulSchema(Schema):
    class Meta:
        name = "delete-success"

    # Successful Delete
    message = fields.Str(required=True)


class LoginInputSchema(Schema):
    """
    Schema for the login route, this is used to validate the data sent to the route.
    """

    class Meta:
        dump = False
        name = "login"

    email = fields.Str(
        required=True,
        load_only=True,
        metadata={
            "description": "Users email address",
            "type": "string",
            "format": "email",
        },
    )
    password = fields.Str(
        required=True,
        load_only=True,
        metadata={
            "description": "Users password",
            "type": "string",
            "format": "password",
        },
    )


class ResetPasswordSchemaIn(Schema):
    class Meta:
        name = "reset-password"

    email = fields.Str(
        required=True,
        load_only=True,
        metadata={
            "description": "Users email address",
            "type": "string",
            "format": "email",
        },
    )


class ResetPasswordSchemaOut(Schema):
    class Meta:
        name = "reset-password-success"

    message = fields.Str(
        dump_only=True,
        metadata={
            "description": "Repsonse message",
            "type": "string",
            "format": "email",
        },
    )


class UserSchema(AutoSchema):
    class Meta:
        name = "user"
        model = User

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @pre_load
    def remove_unwanted_fields(self, data, **kwargs):
        if "password_hash" in data:
            data.pop("password_hash", None)
        if "roles" in data:
            data.pop("roles", None)
        if "password" in data:
            self.password = data.pop("password", None)

        return data

    password = fields.Str(required=False)

    @post_load
    def make_instance(self, data, **kwargs):
        """
        Creates an instance of the model from the data, and sets the password if it exists (there is an issue with
        setter methods and SQLAlchemyAutoSchema, so this is a workaround)
        Args:
            data (dict): The data to be loaded
            **kwargs (dict): The keyword arguments

        Returns:

        """
        instance = self.Meta.model(**data)

        for key, value in data.items():
            setattr(instance, key, value)

        if hasattr(self, "password"):
            instance.password = self.password

        return instance


class TokenRefreshSchema(Schema):
    class Meta:
        name = "token-refresh"

    access_token = fields.Str(
        required=True,
        dump_only=True,
        metadata={
            "description": "Access token",
            "type": "string",
            "format": "password",
        },
    )

    refresh_token = fields.Str(
        required=True,
        dump_only=True,
        metadata={
            "description": "Refresh token",
            "type": "string",
            "format": "password",
        },
    )
    user = fields.Nested(UserSchema, dump_only=True)


class RefreshInputSchema(Schema):
    class Meta:
        dump = False
        name = "refresh-token"

    refresh_token = fields.Str(
        required=True,
        load_only=True,
        metadata={
            "description": "Refresh token",
            "type": "string",
            "format": "password",
        },
    )


class RefreshOutputSchema(Schema):
    class Meta:
        name = "refresh-token-success"

    access_token = fields.Str(
        required=True,
        dump_only=True,
        metadata={
            "description": "Access token",
            "type": "string",
            "format": "password",
        },
    )



import pytest

from flarchitect.utils.general import validate_flask_limiter_rate_limit_string, pluralize_last_word, xml_to_dict
from flarchitect.utils.core_utils import convert_case, dict_to_xml


class TestValidateRateLimitString:

    #  Valid rate limit string with integer and "per" followed by allowed time unit
    def test_valid_rate_limit_string(self):
        # Arrange
        rate_limit_str = "10 per minute"

        # Act
        result = validate_flask_limiter_rate_limit_string(rate_limit_str)
        # Arrange
        rate_limit_str_two = "10 per 5 minutes"

        # Act
        result_two = validate_flask_limiter_rate_limit_string(rate_limit_str_two)

        # Assert
        assert result is True
        assert result_two is True

    #  Empty string
    def test_empty_string(self):
        # Arrange
        rate_limit_str = ""

        # Act
        result = validate_flask_limiter_rate_limit_string(rate_limit_str)

        # Assert
        assert result == False


class TestConvertCase:

    #  The function correctly converts a string to camelCase.
    def test_convert_to_camel_case(self):
        # Arrange
        s = "hello_world"
        target_case = "camel"
        expected_result = "helloWorld"

        # Act
        result = convert_case(s, target_case)

        # Assert
        assert result == expected_result

    #  The function correctly converts a string to PascalCase.
    def test_convert_to_pascal_case(self):
        # Arrange
        s = "hello_world"
        target_case = "pascal"
        expected_result = "HelloWorld"

        # Act
        result = convert_case(s, target_case)

        # Assert
        assert result == expected_result

    #  The function correctly converts a string to snake_case.
    def test_convert_to_snake_case(self):
        # Arrange
        s = "HelloWorld"
        target_case = "snake"
        expected_result = "hello_world"

        # Act
        result = convert_case(s, target_case)

        # Assert
        assert result == expected_result

    #  The function correctly converts a string to SCREAMING_SNAKE_CASE.
    def test_convert_to_screaming_snake_case(self):
        # Arrange
        s = "HelloWorld"
        target_case = "screaming_snake"
        expected_result = "HELLO_WORLD"

        # Act
        result = convert_case(s, target_case)

        # Assert
        assert result == expected_result

    #  The function correctly handles an empty string.
    def test_handle_empty_string(self):
        # Arrange
        s = ""
        target_case = "camel"
        expected_result = ""

        # Act
        result = convert_case(s, target_case)

        # Assert
        assert result == expected_result

    #  The function correctly handles a string with only one word.
    def test_handle_single_word_string(self):
        # Arrange
        s = "hello"
        target_case = "pascal"
        expected_result = "Hello"

        # Act
        result = convert_case(s, target_case)

        # Assert
        assert result == expected_result

    #  The function correctly handles a string with only uppercase letters.
    def test_handle_uppercase_string(self):
        # Arrange
        s = "HELLO_WORLD"
        target_case = "snake"
        expected_result = "hello_world"

        # Act
        result = convert_case(s, target_case)

        # Assert
        assert result == expected_result

    #  The function correctly handles a string with only lowercase letters.
    def test_handle_lowercase_string(self):
        # Arrange
        s = "hello_world"
        target_case = "screaming_snake"
        expected_result = "HELLO_WORLD"

        # Act
        result = convert_case(s, target_case)

        # Assert
        assert result == expected_result



class TestPluralizeLastWord:

    #  pluralizes last word in camelCase
    def test_pluralize_last_word_camel_case(self):
        # Arrange
        converted_name = "camelCase"

        # Act
        result = pluralize_last_word(converted_name)

        # Assert
        assert result == "camelCases"

    #  pluralizes last word in PascalCase
    def test_pluralize_last_word_pascal_case(self):
        # Arrange
        converted_name = "PascalCase"

        # Act
        result = pluralize_last_word(converted_name)

        # Assert
        assert result == "PascalCases"

    #  pluralizes last word in snake_case
    def test_pluralize_last_word_snake_case(self):
        # Arrange
        converted_name = "snake_case"

        # Act
        result = pluralize_last_word(converted_name)

        # Assert
        assert result == "snake_cases"

    #  pluralizes last word in SCREAMING_SNAKE_CASE
    def test_pluralize_last_word_screaming_snake_case(self):
        # Arrange
        converted_name = "SCREAMING_SNAKE_CASE"

        # Act
        result = pluralize_last_word(converted_name)

        # Assert
        assert result == "SCREAMING_SNAKE_CASES"


    #  input with only one word returns the plural of that word
    def test_pluralize_last_word_single_word(self):
        # Arrange
        converted_name = "word"

        # Act
        result = pluralize_last_word(converted_name)

        # Assert
        assert result == "words"

    #  input with only non-alphabetic characters returns the same input


class TestXml:

    def test_simple_dict(self):
        # Arrange
        input_dict = {
            "person": {
                "name": "John",
                "age": 30
            }
        }
        expected_xml = '<?xml version="1.0" encoding="UTF-8" ?><root><person><name>John</name><age>30</age></person></root>'

        # Act
        result = dict_to_xml(input_dict)

        # Assert
        assert result == expected_xml


    def test_nested_dict(self):
        # Arrange
        input_dict = {
            "person": {
                "name": "John",
                "address": {
                    "city": "New York",
                    "zip": "10001"
                }
            }
        }
        expected_xml = '<?xml version="1.0" encoding="UTF-8" ?><root><person><name>John</name><address><city>New York</city><zip>10001</zip></address></person></root>'

        # Act
        result = dict_to_xml(input_dict)

        # Assert
        assert result == expected_xml


    def test_list_in_dict(self):
        # Arrange
        input_dict = {
            "fruits": {
                "fruit": ["apple", "banana", "cherry"]
            }
        }
        expected_xml = '<?xml version="1.0" encoding="UTF-8" ?><root><fruits><fruit><item>apple</item><item>banana</item><item>cherry</item></fruit></fruits></root>'

        # Act
        result = dict_to_xml(input_dict)

        # Assert
        assert result == expected_xml


    def test_root_tag_for_multiple_keys(self):
        # Arrange
        input_dict = {
            "name": "John",
            "age": 30
        }
        expected_xml = '<?xml version="1.0" encoding="UTF-8" ?><root><name>John</name><age>30</age></root>'

        # Act
        result = dict_to_xml(input_dict)

        # Assert
        assert result == expected_xml


class TestXmlDict:
    def test_simple_xml_to_dict(self):
        # Arrange
        xml_data = """<?xml version='1.0' encoding='UTF-8' standalone='no'?><person><name>John</name><age>30</age></person>"""
        expected_dict = {
            "person": {
                "name": "John",
                "age": "30"
            }
        }

        # Act
        result = xml_to_dict(xml_data)

        # Assert
        assert result == expected_dict


    def test_nested_xml_to_dict(self):
        # Arrange
        xml_data = """<?xml version='1.0' encoding='UTF-8' standalone='no'?><person><name>John</name><address><city>New York</city><zip>10001</zip></address></person>"""
        expected_dict = {
            "person": {
                "name": "John",
                "address": {
                    "city": "New York",
                    "zip": "10001"
                }
            }
        }

        # Act
        result = xml_to_dict(xml_data)

        # Assert
        assert result == expected_dict


    def test_xml_with_list_to_dict(self):
        # Arrange
        xml_data = """<?xml version='1.0' encoding='UTF-8' standalone='no'?><fruits><fruit>apple</fruit><fruit>banana</fruit><fruit>cherry</fruit></fruits>"""
        expected_dict = {
            "fruits": {
                "fruit": ["apple", "banana", "cherry"]
            }
        }

        # Act
        result = xml_to_dict(xml_data)

        # Assert
        assert result == expected_dict


    def test_xml_with_empty_elements(self):
        # Arrange
        xml_data = """<?xml version='1.0' encoding='UTF-8' standalone='no'?><person><name>John</name><nickname/><age>30</age></person>"""
        expected_dict = {
            "person": {
                "name": "John",
                "nickname": None,
                "age": "30"
            }
        }

        # Act
        result = xml_to_dict(xml_data)

        # Assert
        assert result == expected_dict


    def test_invalid_xml(self):
        # Arrange
        invalid_xml = """<person><name>John</name><age>30</person>"""  # Malformed XML (unmatched tags)

        # Act & Assert
        with pytest.raises(ValueError, match="Invalid XML data provided"):
            xml_to_dict(invalid_xml)


    def test_bytes_input_xml(self):
        # Arrange
        xml_data = b"""<?xml version='1.0' encoding='UTF-8' standalone='no'?><person><name>John</name><age>30</age></person>"""
        expected_dict = {
            "person": {
                "name": "John",
                "age": "30"
            }
        }

        # Act
        result = xml_to_dict(xml_data)

        # Assert
        assert result == expected_dict



import traceback
from http import HTTPStatus
from typing import Optional, Any

from flask import Response, request
from werkzeug.exceptions import HTTPException

from flarchitect.utils.config_helpers import get_config_or_model_meta
from flarchitect.utils.response_helpers import create_response


class CustomHTTPException(Exception):
    """
    Custom HTTP Exception class
    """

    status_code = None
    error = None
    reason = None

    def __init__(self, status_code, reason=None):
        """
        A custom HTTP exception class

        Args:
            status_code (int): HTTP status code
            reason (str): Reason for the HTTP status code
        """
        self.status_code = status_code
        self.error = HTTPStatus(status_code).phrase  # Fetch the standard HTTP status phrase
        self.reason = reason or self.error  # Use the reason if provided, otherwise use the standard HTTP status phrase

    def to_dict(self):
        return {
            "status_code": self.status_code,
            "status_text": self.error,
            "reason": self.reason,
        }

def handle_http_exception(e: HTTPException) -> Response:
    """
    Handles HTTP exceptions and returns a standardized response.

    Args:
        e (HTTPException): The HTTP exception instance.

    Returns:
        Response: A standardized response object.
    """
    if get_config_or_model_meta(key="API_PRINT_EXCEPTIONS", default=True):
        _print_exception(e)

    prefix = get_config_or_model_meta("API_PREFIX", default="/api")
    if request.path.startswith(prefix):
        return create_response(
            status=e.code, errors={"error": e.name, "reason": e.description}
        )

    # If not an API route, re-raise the exception to let Flask handle it
    return e

def _print_exception(e: Exception) -> None:
    """
    Prints the exception message and stack trace if configured to do so.

    Args:
        e (Exception): The exception to print.
    """
    print(e)
    traceback.print_exc()


def _handle_exception(error: str, status_code: int, error_name: Optional[str] = None, print_exc: bool = True) -> Any:
    """Handles exceptions and formats them into a standardized response."""
    if print_exc:
        import traceback
        traceback.print_exc()

    return create_response(
        status=status_code,
        errors={"error": error, "reason": error_name} # changed this in an attempt to make errors more uniform ##[{"status_text": error_name or "Error", "error": error}]
    )



import time
from typing import Optional, Union, List, Any, Dict, Type

from flask import g
from marshmallow import Schema, ValidationError

from flarchitect.schemas.utils import dump_schema_if_exists
from flarchitect.utils.core_utils import get_count


class CustomResponse:
    """
    Custom response class to be used for serializing output.
    """
    #todo Not really sure why this is here anymore - needs really looking at. Was a use for it once, but no loger it seems.

    def __init__(
        self,
        value: Optional[Union[List, Any]] = None,
        count: Optional[int] = 1,
        error: Optional[Union[List, Dict, Any]] = None,
        status_code: Optional[int] = 200,
        next_url: Optional[str] = None,
        previous_url: Optional[str] = None,
        many: Optional[bool] = False,
        response_ms: Optional[float] = None,
    ):
        self.response_ms = response_ms
        self.value = value
        self.count = count
        self.error = error
        self.status_code = status_code
        self.next_url = next_url
        self.previous_url = previous_url
        self.many = many


def serialize_output_with_mallow(
    output_schema: Type[Schema], data: Any
) -> CustomResponse:
    """
    Utility function to serialize output using a given Marshmallow schema.

    Args:
        output_schema (Type[Schema]): The Marshmallow schema to be used for serialization.
        data (Any): The data to be serialized.

    Returns:
        CustomResponse: The serialized data wrapped in a CustomResponse object.
    """

    try:
        is_list = isinstance(data, list) or (isinstance(data, dict) and ("value" in data or ("query" in data and isinstance(data["query"], list))))
        dump_data = data.get("query", data) if isinstance(data, dict) else data
        value = dump_schema_if_exists(output_schema, dump_data, is_list)
        count = get_count(data, value)

        #Added this is the create_response function as errors were missing the response time
        # response_ms = (time.time() - g.start_time) * 1000 if g.get("start_time") else "n/a"

        return CustomResponse(
            value=value,
            count=count,
            next_url= data.get("next_url") if isinstance(data, dict) else None,
            previous_url=data.get("previous_url") if isinstance(data, dict) else None,
            # response_ms=response_ms,
            many=is_list,
        )

    except ValidationError as err:
        return CustomResponse(value=None, count=None, error=err.messages, status_code=500)


def check_serialise_method_and_return(result: Dict, schema: "AutoSchema", model_columns: List[str], schema_columns: List[str]) -> Union[List[Dict], Any]:
    """
    Checks if the serialization matches the schema or model columns. If not, returns the raw result.

    Args:
        result (Dict): The result dictionary.
        schema (AutoSchema): The schema used for serialization.
        model_columns (List[str]): The model columns.
        schema_columns (List[str]): The schema columns.

    Returns:
        Union[List[Dict], Any]: Serialized data or the original result.
    """
    output_list = result.pop("dictionary", [])
    if output_list:
        output_keys = list(output_list[0].keys())
        if any(x not in model_columns for x in output_keys) or any(x not in schema_columns for x in output_keys):
            return output_list

    return serialize_output_with_mallow(schema, result)



from typing import Dict, Any

from flarchitect.utils.config_helpers import get_config_or_model_meta


def _filter_response_data(data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Filter response data based on configuration settings.

    Args:
        data (Dict[str, Any]): The response data to be filtered.

    Returns:
        Dict[str, Any]: The filtered response data.
    """
    filters = {
        "datetime": "API_DUMP_DATETIME",
        "api_version": "API_DUMP_VERSION",
        "status_code": "API_DUMP_STATUS_CODE",
        "response_ms": "API_DUMP_RESPONSE_MS",
        "total_count": "API_DUMP_TOTAL_COUNT",
    }

    for key, config_key in filters.items():
        if key in data and not get_config_or_model_meta(config_key, default=True):
            data.pop(key)

    for key in ["next_url", "previous_url"]:
        if key in data and not data[key] and not get_config_or_model_meta(f"API_DUMP_NULL_{key.upper()}", default=True):
            data.pop(key)

    if "errors" in data and not data.get("errors") and not get_config_or_model_meta("API_DUMP_NULL_ERRORS", default=False):
        data.pop("errors")

    return data



# Schema for loading login data (username and password)
from marshmallow import Schema, validates, ValidationError
from marshmallow import fields


class LoginSchema(Schema):
    username = fields.String(required=True)
    password = fields.String(required=True)

    # Optional: Add custom validations for username or password, if needed
    @validates("username")
    def validate_username(self, value):
        if not value:
            raise ValidationError("Username is required.")
        if len(value) < 3:
            raise ValidationError("Username must be at least 3 characters long.")

    @validates("password")
    def validate_password(self, value):
        if not value:
            raise ValidationError("Password is required.")
        if len(value) < 6:
            raise ValidationError("Password must be at least 6 characters long.")

# Schema for dumping tokens (auth and refresh tokens)
class TokenSchema(Schema):
    access_token = fields.String(required=True)
    refresh_token = fields.String(required=True)
    user_pk = fields.String(required=True)

# Schema for dumping tokens (auth and refresh tokens)
class RefreshSchema(Schema):
    refresh_token = fields.String(required=True)



import os
import pprint
import re
import socket
from typing import Any, Optional, Dict, Union, Tuple
from xml.etree import ElementTree as ET

import inflect
from flask import Flask, current_app
from jinja2 import Environment, FileSystemLoader

from flarchitect.utils.config_helpers import get_config_or_model_meta
from flarchitect.utils.core_utils import convert_case, get_count
from flarchitect.utils.responses import CustomResponse

HTTP_METHODS = ["GET", "POST", "PATCH", "DELETE"]
DATE_FORMAT = "%Y-%m-%d"
DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
html_path = ""
p = inflect.engine()


class AttributeInitializerMixin:
    """Used for the base Architect flask extension and others to initialize their config attributes."""

    def __init__(self, app: Flask, *args, **kwargs):
        self._set_class_attributes(**kwargs)
        self._set_app_config_attributes(app)
        super().__init__()

    def _set_app_config_attributes(self, app: Flask) -> None:
        """
        Sets class attributes from Flask app config if they exist.

        Args:
            app (Flask): The Flask application instance.
        """
        for key in vars(type(self)).keys():
            if key.startswith("__"):
                continue
            config_key = key.upper().lstrip("_")
            if config_key in app.config:
                setattr(self, key, app.config[config_key])

    def _set_class_attributes(self, **kwargs: Any) -> None:
        """
        Sets class attributes from keyword arguments.

        Args:
            **kwargs: Keyword arguments representing class attributes.
        """
        for key in vars(type(self)).keys():
            if key.startswith("__"):
                continue
            if key in kwargs:
                setattr(self, key, kwargs[key])



def find_html_directory(starting_directory=None):
    # If no starting directory is provided, use the directory of the current file
    if starting_directory is None:
        starting_directory = os.path.abspath(os.path.dirname(__file__))

    # Get the list of all files and directories in the current directory
    contents = os.listdir(starting_directory)

    # Check if "html" directory exists in the current directory
    if 'html' in contents and os.path.isdir(os.path.join(starting_directory, 'html')):
        return os.path.join(starting_directory, 'html')

    # If the current directory is the root directory, stop the search
    parent_directory = os.path.dirname(starting_directory)
    if starting_directory == parent_directory:
        return None

    # Recursively go one level up
    return find_html_directory(parent_directory)

def manual_render_absolute_template(absolute_template_path: str, **kwargs: Any) -> str:
    """
    Manually renders a Jinja2 template given an absolute path.

    Args:
        absolute_template_path (str): The absolute path to the template.
        **kwargs: Additional keyword arguments to pass to the template.

    Returns:
        str: The rendered template as a string.
    """

    template_folder = os.path.join(find_html_directory(), absolute_template_path)
    if template_folder.endswith(".html"):
        template_folder, template_filename = os.path.split(template_folder)

    env = Environment(loader=FileSystemLoader(template_folder))
    template = env.get_template(template_filename)
    return template.render(**kwargs)


def find_child_from_parent_dir(parent: str, child: str, current_dir: str = os.getcwd()) -> Optional[str]:
    """
    Finds the directory of a child folder within a parent directory.

    Args:
        parent (str): The name of the parent directory.
        child (str): The name of the child directory.
        current_dir (str, optional): The current directory to start the search from.

    Returns:
        Optional[str]: The path to the child directory, or None if not found.
    """
    if os.path.basename(current_dir) == parent:
        for dirname in os.listdir(current_dir):
            if dirname == child:
                return os.path.join(current_dir, dirname)

    for dirname in os.listdir(current_dir):
        if dirname.startswith(".") or dirname == "node_modules":
            continue
        child_dir_path = os.path.join(current_dir, dirname)
        if os.path.isdir(child_dir_path):
            child_dir_path = find_child_from_parent_dir(parent, child, child_dir_path)
            if child_dir_path is not None:
                return child_dir_path

    return None


def check_rate_prerequisites(service: str) -> None:
    """
    Checks if prerequisites for a specific service (e.g., Memcached, Redis) are installed.

    Args:
        service (str): The service to check prerequisites for.

    Raises:
        ImportError: If the prerequisite is not available.
    """
    back_end_spec = "or specify a cache service URI in the flask configuration with the key API_RATE_LIMIT_STORAGE_URI={URL}:{PORT}"
    if service == 'Memcached':
        try:
            import pymemcache
        except ImportError:
            raise ImportError("Memcached prerequisite not available. Please install pymemcache " + back_end_spec)
    elif service == 'Redis':
        try:
            import redis
        except ImportError:
            raise ImportError("Redis prerequisite not available. Please install redis-py " + back_end_spec)
    elif service == 'MongoDB':
        try:
            import pymongo
        except ImportError:
            raise ImportError("MongoDB prerequisite not available. Please install pymongo " + back_end_spec)


def check_rate_services() -> Optional[str]:
    """
    Checks if any supported services (Memcached, Redis, MongoDB) are running locally and returns their URI.

    Returns:
        Optional[str]: The URI of the running service, or None if no service is found.
    """
    services = {
        'Memcached': 11211,
        'Redis': 6379,
        'MongoDB': 27017,
    }
    uri = get_config_or_model_meta("API_RATE_LIMIT_STORAGE_URI", default=None)
    if uri:
        return uri

    for service, port in services.items():
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(1)
        try:
            s.connect(('127.0.0.1', port))
            s.close()
            check_rate_prerequisites(service)
            if service == 'Memcached':
                return f'memcached://127.0.0.1:{port}'
            elif service == 'Redis':
                return f'redis://127.0.0.1:{port}'
            elif service == 'MongoDB':
                return f'mongodb://127.0.0.1:{port}'
        except socket.error:
            continue

    return None


def validate_flask_limiter_rate_limit_string(rate_limit_str: str) -> bool:
    """
    Validates a rate limit string for Flask-Limiter.

    Args:
        rate_limit_str (str): The rate limit string to validate.

    Returns:
        bool: True if the rate limit string is valid, False otherwise.
    """
    pattern = re.compile(r'^\d+\s+per\s+(\d+\s+)?(second|minute|hour|day|seconds|minutes|hours|days)$', re.IGNORECASE)
    return bool(pattern.match(rate_limit_str))


def search_all_keys(model: Any, key: str) -> bool:
    """Search for a specific key in all subclasses of a given model.

    Args:
        model (Any): The model class or instance to search in.
        key (str): The key to search for.

    Returns:
        bool: True if the key is found in any subclass, False otherwise.
    """
    for subclass in model.__subclasses__():
        if any(get_config_or_model_meta(key, model=subclass, method=method) for method in HTTP_METHODS):
            return True
    return False


def generate_readme_html(file_path: str, *args: Any, **kwargs: Any) -> str:
    """Generate README content from a Jinja2 template.

    Args:
        file_path (str): The path to the Jinja2 template file.
        *args (Any): Variable length argument list.
        **kwargs (Any): Arbitrary keyword arguments.

    Returns:
        str: The rendered content as a string.
    """
    template_dir, template_file = os.path.split(file_path)
    environment = Environment(loader=FileSystemLoader(os.path.abspath(template_dir)))
    template = environment.get_template(template_file)
    return template.render(*args, **kwargs)


def read_file_content(path: str) -> str:
    """Get the content of a file.

    Args:
        path (str): The path to the file.

    Returns:
        str: The content of the file.

    Raises:
        FileNotFoundError: If the file does not exist.
    """
    if os.path.exists(path):
        with open(path, "r") as file:
            return file.read()
    raise FileNotFoundError(f"{path} not found.")


def case_no_change(s: str) -> str:
    """Return the input string unchanged.

    Args:
        s (str): The input string.

    Returns:
        str: The same input string.
    """
    return s


def pretty_print_dict(d: Dict[Any, Any]) -> str:
    """Pretty print a dictionary.

    Args:
        d (Dict[Any, Any]): The dictionary to pretty print.

    Returns:
        str: The pretty-printed dictionary.
    """
    return pprint.pformat(d, indent=2)


def update_dict_if_flag_true(
    output: Dict[str, Any], flag: bool, key: str, value: Any, case_func: Any
) -> None:
    """Update a dictionary with a key-value pair if the flag is True.

    Args:
        output (Dict[str, Any]): The dictionary to update.
        flag (bool): The flag that controls whether to update.
        key (str): The key to update in the dictionary.
        value (Any): The value to associate with the key.
        case_func (Any): The function to convert the case of the key.
    """
    if flag:
        output.update({convert_case(key, case_func): value})


def make_base_dict() -> Dict[str, Any]:
    """Create a base dictionary with configuration settings.

    Returns:
        Dict[str, Any]: The base dictionary with configuration settings.
    """
    output = {"value": "..."}
    field_case = get_config_or_model_meta("API_FIELD_CASE", default="snake_case")

    config_options = [
        ("API_DUMP_DATETIME", "datetime", "2024-01-01T00:00:00.0000+00:00"),
        ("API_DUMP_VERSION", "api_version", get_config_or_model_meta("API_VERSION", default=True)),
        ("API_DUMP_STATUS_CODE", "status_code", 200),
        ("API_DUMP_RESPONSE_TIME", "response_ms", 15),
        ("API_DUMP_COUNT", "total_count", 10),
        ("API_DUMP_NULL_NEXT_URL", "next_url", "/api/example/url"),
        ("API_DUMP_NULL_PREVIOUS_URL", "previous_url", "null"),
        ("API_DUMP_NULL_ERRORS", "errors", "null", False),
    ]

    for config, key, value, *defaults in config_options:
        flag = get_config_or_model_meta(config, default=defaults[0] if defaults else True)
        update_dict_if_flag_true(output, flag, key, value, field_case)

    return output


def pluralize_last_word(converted_name: str) -> str:
    """
    Pluralize the last word of the converted name while preserving the rest of the name and its case.

    Args:
        converted_name (str): The name after case conversion.

    Returns:
        str: The name with the last word pluralized.
    """
    delimiters = {"_": "snake", "-": "kebab"}
    delimiter = next((d for d in delimiters if d in converted_name), "")

    words = converted_name.split(delimiter) if delimiter else re.findall(r'[A-Z]?[a-z]+|[A-Z]+(?![a-z])', converted_name)
    last_word = words[-1]
    last_word_pluralized = p.plural(p.singular_noun(last_word) or last_word)

    words[-1] = last_word_pluralized
    pluralized_name = delimiter.join(words)

    if delimiters.get(delimiter) in ["screaming_snake", "screaming_kebab"]:
        pluralized_name = pluralized_name.upper()

    return pluralized_name


def normalize_key(key: str) -> str:
    """
    Converts a key to uppercase.

    Args:
        key (str): The key to be normalized.

    Returns:
        str: The normalized key.
    """
    return key.upper()

def xml_to_dict(xml_data: Union[str, bytes]) -> Dict[str, Any]:
    """
    Converts an XML string or bytes into a dictionary.
    Args:
        xml_data (Union[str, bytes]): The XML data.
    Returns:
        Dict[str, Any]: The resulting dictionary.
    """
    xml_data = xml_data.decode() if hasattr(xml_data, "decode") else xml_data

    def element_to_dict(element: ET.Element) -> Any:
        if not list(element) and (element.text is None or not element.text.strip()):
            return None
        if element.text and element.text.strip() and not list(element):
            return element.text.strip()
        result = {}
        for child in element:
            child_result = element_to_dict(child)
            if child.tag not in result:
                result[child.tag] = child_result
            else:
                if not isinstance(result[child.tag], list):
                    result[child.tag] = [result[child.tag]]
                result[child.tag].append(child_result)
        return result

    try:
        root = ET.fromstring(xml_data)
    except ET.ParseError as e:
        raise ValueError("Invalid XML data provided") from e
    return {root.tag: element_to_dict(root)}


def handle_result(result: Any) -> Tuple[int, Any, int, Optional[str], Optional[str]]:
    """Processes the result of a route function and prepares it for the standardized response."""

    #todo really not sure why this is here again. Its a relic from the past, a lot of this needs looking at.

    status_code, value, count, next_url, previous_url = HTTP_OK, result, 1, None, None

    if isinstance(result, tuple):
        status_code, result = (result[1], result[0]) if len(result) == 2 and isinstance(result[1], int) else (HTTP_OK, result)
    if isinstance(result, dict):
        value, count = result.get("query", result), get_count(result, result.get("query"))
        next_url, previous_url = result.get("next_url"), result.get("previous_url")
    elif isinstance(result, CustomResponse):
        next_url, previous_url, count = result.next_url, result.previous_url, result.count

    return status_code, value, count, next_url, previous_url


HTTP_OK = 200
HTTP_BAD_REQUEST = 400
HTTP_INTERNAL_SERVER_ERROR = 500
HTTP_UNAUTHORIZED = 401
HTTP_FORBIDDEN = 403



import re
from typing import Dict, Any
from xml.dom.expatbuilder import parseString
from xml.etree import ElementTree as ET

from dicttoxml import dicttoxml
import xml.dom.minidom


def convert_case(s: str, target_case: str) -> str:
    """
    Convert a string to the specified case format.

    Args:
        s (str): The input string.
        target_case (str): The target case format (camel, pascal, snake, screaming_snake, kebab, screaming_kebab).

    Returns:
        str: The converted string.
    """
    # Handle empty strings
    if not s:
        return s

    # Regex to handle acronyms and camelCase/PascalCase
    words = re.findall(r"[A-Z]+(?=[A-Z][a-z])|[A-Z]?[a-z]+|[A-Z]+|[0-9]+", s)
    words = [word.lower() for word in words]

    # Handle cases where `words` might be empty
    if not words:
        return s

    # Define case converters
    case_converters = {
        "camel": lambda: words[0] + "".join(word.capitalize() for word in words[1:]),
        "pascal": lambda: "".join(word.capitalize() for word in words),
        "snake": lambda: "_".join(words),
        "screaming_snake": lambda: "_".join(word.upper() for word in words),
        "kebab": lambda: "-".join(words),
        "screaming_kebab": lambda: "-".join(word.upper() for word in words),
    }

    # Convert and return the case format
    return case_converters.get(target_case, lambda: s)()


def convert_snake_to_camel(snake_str: str) -> str:
    """Convert a snake_case string to camelCase, preserving a leading underscore if present."""
    leading_underscore = snake_str.startswith("_")
    camel_case_str = convert_case(snake_str, "camel")
    return ("_" if leading_underscore else "") + camel_case_str


def convert_camel_to_snake(camel_str: str) -> str:
    """Convert a camelCase string to snake_case."""
    return convert_case(camel_str, "snake")


def convert_kebab_to_snake(kebab_str: str) -> str:
    """Convert a kebab-case string to snake_case."""
    return convert_case(kebab_str, "snake")


def dict_to_xml(input_dict: dict) -> str:
    # Convert dict to XML bytes
    xml_bytes = dicttoxml(input_dict, custom_root="root", attr_type=False)

    # Pretty print the XML string
    return xml_bytes.decode()


def get_count(result: Dict[str, Any], value: Any) -> int:
    """Determine the count of records in the result."""
    if isinstance(result, dict) and result.get("total_count"):
        return result["total_count"]
    return len(value) if isinstance(value, list) else (0 if not value else 1)



import pytest

from demo.model_extension.model import create_app


@pytest.fixture
def app_model_meta():
    app_model_meta = create_app(
        {
            "API_TITLE": "Automated test",
            "API_VERSION": "0.2.0",
            # Other configurations specific to this test
        }
    )
    yield app_model_meta


@pytest.fixture
def client_meta_model(app_model_meta):
    return app_model_meta.test_client()


def test_model_block_methods(client_meta_model):

    fail_get = client_meta_model.get("/api/categories/1")
    assert fail_get.status_code == 405



import re

from colorama import Fore, Style, init

# Initialize Colorama
init(autoreset=True)


def color_text_with_multiple_patterns(text):
    """Color text wrapped in specific patterns with respective colors."""

    # Patterns to match text: backticks, hyphens, pluses, dollars, and pipes
    patterns = {
        r"`(.*?)`": (Fore.YELLOW, Style.NORMAL),  # Yellow for backticks
        r"\+(.*?)\+": (Fore.RED, Style.NORMAL),  # Red for pluses
        r"-(.*?)-": (Fore.CYAN, Style.NORMAL),  # Cyan for hyphens
        r"\$(.*?)\$": (Fore.MAGENTA, Style.BRIGHT),  # Magenta for dollars
        r"\|(.*?)\|": (Fore.GREEN, Style.BRIGHT),  # Green for pipes
    }

    # Function to replace matched text with colored version
    def replace_with_color(match, color, style):
        return color + style + match.group(1) + Style.RESET_ALL

    # Iterate over the patterns and apply each one
    for pattern, (color, style) in patterns.items():
        # Pass both color and style to the lambda function
        text = re.sub(pattern, lambda match: replace_with_color(match, color, style), text)

    return text


class CustomLogger:
    def __init__(self, verbosity_level=0):
        self.verbosity_level = verbosity_level

    def _log(self, text: str):
        """
        Log a message to the console.
        Args:
            text (str): The message to log.

        Returns:
            None
        """
        print(color_text_with_multiple_patterns(text))

    def log(self, level, message):
        """Log a message if its level is less than or equal to the current verbosity level."""
        if level <= self.verbosity_level:
            self._log(f"LOG {level}: ".ljust(10) + message)

    def debug(self, level, message):
        """Log a message if its level is less than or equal to the current verbosity level."""
        if level <= self.verbosity_level:
            self._log(f"DEBUG {level}: ".ljust(10) + message)

    def error(self, level, message):
        """Log a message if its level is less than or equal to the current verbosity level."""
        if level <= self.verbosity_level:
            self._log(Fore.RED + f"ERROR {level}: ".ljust(10) + Style.RESET_ALL + message)


logger = CustomLogger()



import os
import secrets
import time
from datetime import timedelta
from types import FunctionType
from typing import List, Dict, Any, Callable, Optional, Type, Union

from flask import abort, request, Blueprint, current_app, g
from flask_jwt_extended import (
    create_access_token,
    create_refresh_token,
    get_jwt_identity,
)
from flask_login import login_user, logout_user
from marshmallow import Schema
from sqlalchemy import inspect
from sqlalchemy.orm import DeclarativeBase, Session
from werkzeug.exceptions import default_exceptions, MethodNotAllowed

from flarchitect.authentication.jwt import generate_access_token, generate_refresh_token, refresh_access_token, \
    get_user_from_token, get_pk_and_lookups
from flarchitect.core.utils import get_url_pk, get_foreign_key_to_parent, get_primary_key_info
from flarchitect.database.operations import CrudService
from flarchitect.database.utils import get_models_relationships, get_primary_keys
from flarchitect.exceptions import CustomHTTPException, handle_http_exception
from flarchitect.logging import logger
from flarchitect.schemas.auth import TokenSchema, LoginSchema, RefreshSchema
from flarchitect.schemas.utils import get_input_output_from_model_or_make

from flarchitect.specs.utils import (
    get_param_schema,
    generate_delete_query_params,
    generate_get_query_params,
    generate_additional_query_params,
    get_description,
    get_tag_group,
    endpoint_namer,
)
from flarchitect.utils.general import AttributeInitializerMixin
from flarchitect.utils.config_helpers import get_config_or_model_meta


def create_params_from_rule(
    model: DeclarativeBase, rule, schema: Schema
) -> List[Dict[str, Any]]:
    """Generates path parameters from a Flask routing rule.

    Args:
        model (DeclarativeBase): Model to generate path parameters from.
        rule: Rule to generate path parameters from.
        schema (Schema): The schema associated with the rule.

    Returns:
        List[Dict[str, Any]]: List of path parameters with enhanced type checks and descriptions.
    """
    path_params = []

    for argument in rule.arguments:
        name = get_config_or_model_meta(
            "name", model=model, output_schema=schema, default=None
        )
        if not name:
            name = (
                (model or schema).__name__.replace("Schema", "").replace("schema", "")
            )

        param_info = {
            "name": argument,
            "in": "path",
            "required": True,
            "description": f"Identifier for the {name} instance.",
            "schema": get_param_schema(rule._converters[argument]),
        }

        path_params.append(param_info)

    return path_params


def create_query_params_from_rule(
    rule,
    methods: set,
    schema: Schema,
    many: bool,
    model: DeclarativeBase,
    custom_query_params: List[Dict[str, Any]] = [],
) -> List[Dict[str, Any]]:
    """Generates query parameters from a rule.

    Args:
        rule: Rule to generate query parameters from.
        methods (set): Set of methods to generate query parameters from.
        schema (Schema): Schema to generate query parameters from.
        many (bool): Whether the endpoint returns multiple items.
        model (DeclarativeBase): Model to generate query parameters from.
        custom_query_params (List[Dict[str, Any]]): Custom query parameters to append to the generated query parameters.

    Returns:
        List[Dict[str, Any]]: List of query parameters.
    """
    query_params = (
        generate_delete_query_params(schema, model) if "DELETE" in methods else []
    )

    if "GET" in methods and many:
        query_params.extend(generate_get_query_params(schema, model))

    query_params.extend(generate_additional_query_params(methods, schema, model))

    if custom_query_params:
        query_params.extend(custom_query_params)

    return query_params


def find_rule_by_function(architect, f: Callable):
    """Gets the path, methods, and path parameters for a function.

    Args:
        architect: The architect object.
        f (Callable): The function to get the path, methods, and path parameters for.

    Returns:
        The rule associated with the function.
    """
    for rule in architect.app.url_map.iter_rules():
        if rule.endpoint.split(".")[-1] == f.__name__:
            return rule
    return None


def create_route_function(
    service,
    method: str,
    many: bool,
    join_model: Optional[Type[DeclarativeBase]] = None,
    get_field: Optional[str] = None,
    **kwargs,
) -> Callable:
    """
    Sets up the route function for the API based on the HTTP method.

    Args:
        service: The CRUD service for the model.
        method (str): The HTTP method (GET, POST, PATCH, DELETE).
        many (bool): Whether the route handles multiple records.
        join_model (Optional[Type[DeclarativeBase]]): The model to use in the join.
        get_field (Optional[str]): The field to get the record by.

    Returns:
        Callable: The route function.
    """

    def pre_process(pre_hook: Callable, **hook_kwargs) -> Dict[str, Any]:
        if pre_hook:
            model = hook_kwargs.pop("model", None) or service.model
            return pre_hook(model=model, **hook_kwargs)
        return hook_kwargs

    def post_process(post_hook: Callable, output: Any, **hook_kwargs) -> Any:
        if post_hook:
            model = hook_kwargs.pop("model", None) or service.model
            out_val = post_hook(model=model, output=output, **hook_kwargs).get("output")

            # Ensure out_val is valid before attempting to access "output"
            return (
                out_val.get("output")
                if isinstance(out_val, dict) and "output" in out_val
                else out_val
            )

        return output

    def route_function_factory(
        action: Callable,
        many: bool,
        pre_hook: Optional[Callable],
        post_hook: Optional[Callable],
    ) -> Callable:
        def route_function(id: Optional[int] = None, **hook_kwargs) -> Any:
            hook_kwargs = pre_process(
                pre_hook, id=id, field=get_field, join_model=join_model, **hook_kwargs
            )
            action_kwargs = {"lookup_val": id} if id else {}
            action_kwargs.update(hook_kwargs)
            action_kwargs["many"] = many
            action_kwargs["data_dict"] = hook_kwargs.get("deserialized_data")
            action_kwargs["join_model"] = hook_kwargs.get("join_model")
            action_kwargs["id"] = hook_kwargs.get("id")
            action_kwargs["model"] = hook_kwargs.get("model")

            output = action(**action_kwargs) or abort(404)
            return post_process(post_hook, output, **hook_kwargs)

        return route_function

    pre_hook = get_config_or_model_meta(
        "API_SETUP_CALLBACK", model=service.model, default=None, method=method
    )
    post_hook = get_config_or_model_meta(
        "API_RETURN_CALLBACK", model=service.model, default=None, method=method
    )

    action_map = {
        "GET": lambda **action_kwargs: service.get_query(
            request.args.to_dict(), alt_field=get_field, **action_kwargs
        ),
        "DELETE": service.delete_object,
        "PATCH": service.update_object,
        "POST": service.add_object,
    }

    action = action_map.get(method)
    return route_function_factory(action, many, pre_hook, post_hook)


class RouteCreator(AttributeInitializerMixin):
    created_routes: Dict[str, Dict[str, Any]] = {}
    architect: "Architect"
    api_full_auto: Optional[bool] = True
    api_base_model: Optional[Union[Callable, List[Callable]]] = None
    api_base_schema: Optional[Callable] = None
    db_service: Optional[Callable] = CrudService
    session: Optional[Union[Session, List[Session]]] = None
    blueprint: Optional[Blueprint] = None

    def __init__(self, architect: "Architect", *args, **kwargs):
        """Initialize the RouteCreator object.

        Args:
            architect (Architect): The architect object.
            *args (list): List of arguments.
            **kwargs (dict): Dictionary of keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.architect = architect
        if self.api_full_auto:
            self.setup_models()
            self.validate()
            self.setup_api_routes()

    def setup_models(self):
        """Set up the models for the API by adding necessary configurations."""
        self.api_base_model = (
            [self.api_base_model]
            if not isinstance(self.api_base_model, list)
            else self.api_base_model
        )

        for base in self.api_base_model:
            for model_class in base.__subclasses__():
                # Add any necessary setup here for model_class
                pass

    def validate(self):
        """Validate the RiceAPI configuration."""
        if self.api_full_auto:
            self._validate_base_model_setup()
            self._validate_authentication_setup()
            self._validate_soft_delete_setup()

    def _validate_base_model_setup(self):
        """Validate the base model setup for the API."""
        if not self.api_base_model:
            raise ValueError(
                "If FULL_AUTO is True, API_BASE_MODEL must be set to a SQLAlchemy model."
            )

        self.api_base_model = (
            [self.api_base_model]
            if not isinstance(self.api_base_model, list)
            else self.api_base_model
        )

        for base in self.api_base_model:
            if not hasattr(base, "get_session"):
                raise ValueError(
                    "If FULL_AUTO is True, API_BASE_MODEL must have a `get_session` function that returns"
                    "the database session for that model."
                )

    def _validate_authentication_setup(self):
        """Validate the authentication setup for the API."""
        user = get_config_or_model_meta("API_USER_MODEL", default=None)
        auth = get_config_or_model_meta("API_AUTHENTICATE", default=None)
        auth_method = get_config_or_model_meta("API_AUTHENTICATE_METHOD", default=None)

        if not current_app.config.get("SECRET_KEY") and auth:
            raise ValueError(
                "SECRET_KEY must be set in the Flask app config. You can use this randomly generated key:\n"
                f"{secrets.token_urlsafe(48)}\n"
                f"And this SALT key\n"
                f"{secrets.token_urlsafe(32)}\n"
            )

        if auth and not user and callable(auth):
            raise ValueError(
                "If API_AUTHENTICATE is set to a callable, API_USER_MODEL must be set to the user model."
            )

        if auth and not auth_method:
            raise ValueError(
                "If API_AUTHENTICATE is set to True, API_AUTHENTICATE_METHOD must be set to either 'basic', 'jwt', or 'api_key'."
            )

        if auth_method == "jwt":
            ACCESS_SECRET_KEY = os.environ.get("ACCESS_SECRET_KEY") or self.architect.app.config.get(
                "ACCESS_SECRET_KEY"
            )
            REFRESH_SECRET_KEY = os.environ.get("REFRESH_SECRET_KEY") or self.architect.app.config.get(
                "REFRESH_SECRET_KEY"
            )
            if not ACCESS_SECRET_KEY or not REFRESH_SECRET_KEY:
                raise ValueError(
                    """If API_AUTHENTICATE_METHOD is set to 'jwt' you must set ACCESS_SECRET_KEY and REFRESH_SECRET_KEY in the
                    Flask app config or as environment variables.""")

    def _validate_soft_delete_setup(self):
        """Validate the soft delete setup for the API."""
        soft_delete = get_config_or_model_meta("API_SOFT_DELETE", default=False)
        if soft_delete:
            deleted_attr = get_config_or_model_meta(
                "API_SOFT_DELETE_ATTRIBUTE", default=None
            )
            soft_delete_values = get_config_or_model_meta(
                "API_SOFT_DELETE_VALUES", default=None
            )

            if not deleted_attr:
                raise ValueError(
                    "If API_SOFT_DELETE is set to True, API_SOFT_DELETE_ATTRIBUTE must be set to the name of the "
                    "attribute that holds the soft delete value."
                )

            if (
                not soft_delete_values
                or not isinstance(soft_delete_values, tuple)
                or len(soft_delete_values) != 2
            ):
                raise ValueError(
                    "API_SOFT_DELETE_VALUES must be a tuple of two values that represent the soft delete state (not deleted, deleted)."
                )

    def setup_api_routes(self):
        """Setup all necessary API routes."""
        self.make_auth_routes()
        self.create_api_blueprint()
        self.create_routes()
        self.make_exception_routes()
        self.architect.app.register_blueprint(self.blueprint)

    def make_exception_routes(self):
        for code in default_exceptions.keys():
            logger.debug(
                4,
                f"Setting up custom error handler for blueprint |{self.blueprint.name}| with http code +{code}+.",
            )
            self.architect.app.register_error_handler(code, handle_http_exception)

    def create_routes(self):
        """Create all the routes for the API."""
        for base in self.api_base_model:
            for model_class in base.__subclasses__():
                if hasattr(model_class, "__table__") and hasattr(model_class, "Meta"):
                    session = model_class.get_session()
                    self.make_all_model_routes(model_class, session)

    def make_auth_routes(self):
        """Create the authentication routes for the API."""
        auth_method = get_config_or_model_meta("API_AUTHENTICATE_METHOD", default=None)
        user = get_config_or_model_meta("API_USER_MODEL", default=None)

        if not auth_method:
            return

        from flask_login import LoginManager

        login_manager = LoginManager()
        login_manager.init_app(self.architect.app)

        if auth_method == "jwt":
            self._make_jwt_auth_routes(user)
        elif auth_method == "basic":
            self._make_basic_auth_routes(user)
        elif auth_method == "api_key":
            self._make_api_key_auth_routes(user)

        @login_manager.user_loader
        def load_user(user_id):
            return user.get(user_id)

    def _make_basic_auth_routes(self, user: Callable):
        """Create basic authentication routes."""
        pass  # Implement basic auth routes here

    def _make_api_key_auth_routes(self, user: Callable):
        """Create API key authentication routes."""
        pass  # Implement API key auth routes here

    def _make_jwt_auth_routes(self, user: Callable):
        """Create JWT authentication routes."""
        self._create_jwt_login_route(user)
        self._create_jwt_logout_route(user)
        self._create_jwt_refresh_route(user)

    def _create_jwt_login_route(self, user: Callable):
        """Create the login route for JWT authentication."""

        @self.architect.app.route("/login", methods=["POST"])
        @self.architect.schema_constructor(
            input_schema=LoginSchema,
            output_schema=TokenSchema,
            model=user,
            many=False,
            roles=True,
            group_tag="Authentication",
        )
        def login(*args, **kwargs):
            data = request.get_json()
            username = data.get("username")
            password = data.get("password")

            hash_field = get_config_or_model_meta("API_PASSWORD_HASH_FIELD", model=user, default=None)
            lookup_field = get_config_or_model_meta("API_USER_LOOKUP_FIELD", model=user, default=None)
            check_method = get_config_or_model_meta("API_PASSWORD_CHECK_METHOD", model=user, default=None)

            usr = user.query.filter(getattr(user, lookup_field)==username).first()

            if usr and getattr(usr, check_method)(password):
                access_token = generate_access_token(usr)
                refresh_token = generate_refresh_token(usr)

                pk, lookup_field = get_pk_and_lookups()

                return {"access_token": access_token, "refresh_token": refresh_token, "user_pk": getattr(usr, pk)}

            raise CustomHTTPException(401, "Invalid credentials")

    def _create_jwt_logout_route(self, user: Callable):
        """Create the logout route for JWT authentication."""

        @self.architect.app.route("/logout", methods=["POST"])
        @self.architect.schema_constructor(
            output_schema=None,
            many=False,
            group_tag="Authentication",
        )
        def logout(*args, **kwargs):
            logout_user()
            return {}

    def _create_jwt_refresh_route(self, *args, **kwargs):

        """Create the refresh token route for JWT authentication."""

        @self.architect.app.route("/refresh", methods=["POST"])
        @self.architect.schema_constructor(
            input_schema=RefreshSchema,
            output_schema=TokenSchema,
            many=False,
            group_tag="Authentication",
        )
        def refresh(*args, **kwargs):
            """
            Endpoint to refresh JWT access tokens using a refresh token.

            Expects:
                JSON payload with 'refresh_token'.

            Returns:
                JSON response with 'access_token' and 'refresh_token'.
            """

            # Extract the refresh token from the request
            refresh_token = request.get_json().get("refresh_token")
            if not refresh_token:
                raise CustomHTTPException(status_code=400, reason='Refresh token is missing')

            # Attempt to refresh the access token and retrieve the user
            try:
                new_access_token, user = refresh_access_token(refresh_token)
            except CustomHTTPException as e:
                # Let your application's error handlers manage the response
                raise e

            # Generate a new refresh token
            new_refresh_token = generate_refresh_token(user)

            pk, lookup_field = get_pk_and_lookups()
            # Return the new tokens
            return {"access_token": new_access_token, "refresh_token": new_refresh_token, "user_pk": getattr(user, pk)}

    def create_api_blueprint(self):
        """Register the API blueprint and error handlers."""
        api_prefix = get_config_or_model_meta("API_PREFIX", default="/api")
        self.blueprint = Blueprint("api", __name__, url_prefix=api_prefix)

        @self.architect.app.before_request
        def before_request(*args, **kwargs):
            g.start_time = time.time()

    def make_all_model_routes(self, model: Callable, session: Any):
        """Create all routes for a given model.

        Args:
            model (Callable): The model to create routes for.
            session (Any): The database session to use for the model.
        """
        self._generate_relation_routes(model, session)
        self._generate_model_routes(model, session)

    def _generate_model_routes(self, model: Callable, session: Any):
        """Generate CRUD routes for a model.

        Args:
            model (Callable): The model to create routes for.
            session (Any): The database session to use for the model.
        """

        # Retrieve allowed and blocked methods from configuration or model metadata

        read_only = get_config_or_model_meta(
            "API_READ_ONLY", model=model, default=False
        )

        allowed, allowed_from = get_config_or_model_meta(
            "API_ALLOWED_METHODS", model=model, default=[], return_from_config=True
        )
        allowed_methods = [x.upper() for x in allowed]

        blocked_methods, blocked_from = get_config_or_model_meta(
            "API_BLOCK_METHODS",
            model=model,
            default=[],
            allow_join=True,
            return_from_config=True,
        )
        blocked_methods = [x.upper() for x in blocked_methods]

        for http_method in ["GETS", "GET", "POST", "PATCH", "DELETE"]:

            if (
                read_only
                and http_method in ["POST", "PATCH", "DELETE"]
                and (http_method not in allowed_methods)
            ):
                continue

            check_http_meth = (
                http_method
                if not http_method.endswith("S")
                else http_method.replace("S", "")
            )
            if (
                check_http_meth in blocked_methods
                and blocked_from == "config"
                and allowed_from == "default"
            ):
                continue

            if check_http_meth not in allowed_methods and allowed_from == "config":
                continue

            if check_http_meth in blocked_methods and blocked_from == "model":
                continue
            if check_http_meth in blocked_methods and (
                not check_http_meth in allowed_methods and allowed_from == "model"
            ):
                continue

            if (
                check_http_meth not in allowed_methods
                and allowed_methods
                and allowed_from in ["config", "default"]
            ):
                continue

            if (
                allowed_methods
                and allowed_from == "model"
                and check_http_meth not in allowed_methods
            ):
                continue

            route_data = self._prepare_route_data(model, session, http_method)
            self.generate_route(**route_data)

    def _generate_relation_routes(self, model: Callable, session: Any):
        """Generate routes for model relationships if configured.

        Args:
            model (Callable): The model to create relation routes for.
            session (Any): The database session to use for the model.
        """
        if get_config_or_model_meta("API_ADD_RELATIONS", model=model, default=True):
            relations = get_models_relationships(model)
            for relation_data in relations:
                prepared_relation_data = self._prepare_relation_route_data(
                    relation_data, session
                )
                self._create_relation_route_and_to_url_function(prepared_relation_data)

    def _create_relation_route_and_to_url_function(self, relation_data: Dict[str, Any]):
        """Create a route for a relation and add a to_url function to the model.

        Args:
            relation_data (Dict[str, Any]): The data for creating the relation route.
        """
        child = relation_data["child_model"]
        parent = relation_data["parent_model"]
        self._add_relation_url_function_to_model(
            child=child, parent=parent, id_key=relation_data["join_key"]
        )
        self.generate_route(**relation_data)

    def _prepare_route_data(
        self, model: Callable, session: Any, http_method: str
    ) -> Dict[str, Any]:
        """Prepare data for creating a route.

        Args:
            model (Callable): The model to create the route for.
            session (Any): The database session to use for the model.
            http_method (str): The HTTP method for the route.

        Returns:
            Dict[str, Any]: The prepared route data.
        """

        many = False
        if http_method == "GETS":
            many = True
            http_method = "GET"

        input_schema_class, output_schema_class = get_input_output_from_model_or_make(
            model
        )

        base_url = f"/{self._get_url_naming_function(model, input_schema_class, output_schema_class)}"
        method = http_method

        if http_method == "GET" and not many or http_method in ["DELETE", "PATCH"]:
            pk_url = get_url_pk(
                model
            )  # GET operates on a single item, so include the primary key in the URL
            base_url = f"{base_url}/{pk_url}"

        logger.debug(
            4,
            f"Collecting main model data for -{model.__name__}- with expected url |{method}|:`{base_url}`.",
        )

        return {
            "model": model,
            "many": many,
            "method": method,
            "url": base_url,
            "name": model.__name__.lower(),
            "output_schema": output_schema_class,
            "session": session,
            "input_schema": (
                input_schema_class if http_method in ["POST", "PATCH"] else None
            ),
        }

    def _prepare_relation_route_data(
        self, relation_data: Dict[str, Any], session: Any
    ) -> Dict[str, Any]:
        """Prepare data for creating a relation route.

        Args:
            relation_data (Dict[str, Any]): Data about the relation.
            session (Any): The database session to use for the relation.

        Returns:
            Dict[str, Any]: The prepared relation route data.
        """

        child_model = relation_data["model"]
        parent_model = relation_data["parent"]
        input_schema_class, output_schema_class = get_input_output_from_model_or_make(
            child_model
        )
        pinput_schema_class, poutput_schema_class = get_input_output_from_model_or_make(
            parent_model
        )

        key = get_primary_key_info(parent_model)

        relation_url = (
            f"/{self._get_url_naming_function(parent_model, pinput_schema_class, poutput_schema_class)}"
            f"/<{key[1]}:{key[0]}>"
            f"/{self._get_url_naming_function(child_model, input_schema_class, output_schema_class)}"
        )
        logger.debug(
            4,
            f"Collecting parent/child model relationship for -{parent_model.__name__}- and -{child_model.__name__}- with expected url `{relation_url}`.",
        )

        return {
            "child_model": child_model,
            "model": child_model,
            "parent_model": parent_model,
            "many": relation_data["join_type"][-4:].lower() == "many"
            or relation_data.get("many", False),
            "method": "GET",
            "relation_name": relation_data["relationship"],
            "url": relation_url,
            "name": f"{child_model.__name__.lower()}_join_to_{parent_model.__name__.lower()}",
            "join_key": relation_data["right_column"],
            "output_schema": output_schema_class,
            "session": session,
        }

    def generate_route(self, **kwargs: Dict[str, Any]):
        """Generate the route for this method/model.

        Args:
            **kwargs (Dict[str, Any]): Dictionary of keyword arguments for route generation.
        """
        description = get_description(kwargs)
        kwargs["group_tag"] = get_tag_group(kwargs)
        model = kwargs.get("model", kwargs.get("child_model"))
        service = CrudService(model=model, session=kwargs["session"])
        http_method = kwargs.get("method", "GET")

        # Ensure the route is not blocked
        # if self._is_route_blocked(http_method, model):
        #     return

        route_function = create_route_function(
            service,
            http_method,
            many=kwargs.get("many", False),
            join_model=kwargs.get("parent_model", None),
            get_field=kwargs.get("join_key"),
        )

        unique_route_function = self._create_unique_route_function(
            route_function, kwargs["url"], http_method, kwargs.get("many", False)
        )
        kwargs["function"] = unique_route_function

        # Register the route with Flask
        self._add_route_to_flask(
            kwargs["url"],
            kwargs["method"],
            self.architect.schema_constructor(**kwargs)(unique_route_function),
        )
        (
            self._add_self_url_function_to_model(model)
            if not kwargs.get("join_key")
            else None
        )
        self._add_to_created_routes(**kwargs)

    def _is_route_blocked(self, http_method: str, model: Callable) -> bool:
        """Check if the route is blocked based on the configuration.

        Args:
            http_method (str): The HTTP method of the route.
            model (Callable): The model for the route.

        Returns:
            bool: True if the route is blocked, otherwise False.
        """
        blocked_methods = get_config_or_model_meta(
            "API_BLOCK_METHODS", model=model, default=[], allow_join=True
        )
        read_only = get_config_or_model_meta(
            "API_READ_ONLY", model=model, default=False
        )
        if read_only:
            blocked_methods.extend(["POST", "PATCH", "DELETE"])

        return http_method in [x.upper() for x in blocked_methods]

    def _create_unique_route_function(
        self,
        route_function: Callable,
        url: str,
        http_method: str,
        is_many: bool = False,
    ) -> Callable:
        """Create a unique route function name.

        Args:
            route_function (Callable): The original route function.
            url (str): The URL of the route.
            http_method (str): The HTTP method of the route.

        Returns:
            Callable: The unique route function.
        """
        # Ensure the function name is unique by differentiating between collection and single item routes
        if is_many:
            unique_function_name = (
                f"route_wrapper_{http_method}_collection_{url.replace('/', '_')}"
            )
        else:
            unique_function_name = (
                f"route_wrapper_{http_method}_single_{url.replace('/', '_')}"
            )

        unique_route_function = FunctionType(
            route_function.__code__,
            globals(),
            unique_function_name,
            route_function.__defaults__,
            route_function.__closure__,
        )
        return unique_route_function

    def _add_route_to_flask(self, url: str, method: str, function: Callable):
        """Add a route to Flask.

        Args:
            url (str): The URL endpoint.
            method (str): The HTTP method.
            function (Callable): The function to call when the route is visited.
        """
        logger.log(1, f"|{method}|:`{self.blueprint.url_prefix}{url}` added to flask.")
        self.blueprint.add_url_rule(url, view_func=function, methods=[method])

    def _add_self_url_function_to_model(self, model: Callable):
        """Add a self URL method to the model class.

        Args:
            model (Callable): The model to add the function to.
        """
        primary_keys = [key.name for key in model.__table__.primary_key]
        if len(primary_keys) > 1:
            logger.error(
                1,
                f"Composite primary keys are not supported, failed to set method $to_url$ on -{model.__name__}-",
            )
            return

        api_prefix = get_config_or_model_meta("API_PREFIX", default="/api")
        url_naming_function = get_config_or_model_meta(
            "API_ENDPOINT_NAMER", model, default=endpoint_namer
        )

        def to_url(self):
            return f"{api_prefix}/{url_naming_function(model)}/{getattr(self, primary_keys[0])}"

        logger.log(3, f"Adding method $to_url$ to model -{model.__name__}-")
        setattr(model, "to_url", to_url)

    def _add_relation_url_function_to_model(
        self, id_key: str, child: Callable, parent: Callable
    ):
        """Add a relation URL method to the model class.

        Args:
            id_key (str): The primary key attribute name.
            child (Callable): The child model.
            parent (Callable): The parent model.
        """
        api_prefix = get_config_or_model_meta("API_PREFIX", default="/api")
        parent_endpoint = get_config_or_model_meta(
            "API_ENDPOINT_NAMER", parent, default=endpoint_namer
        )(parent)
        child_endpoint = get_config_or_model_meta(
            "API_ENDPOINT_NAMER", child, default=endpoint_namer
        )(child)

        def to_url(self):
            parent_pk = get_primary_keys(parent).key
            return f"{api_prefix}/{parent_endpoint}/{getattr(self, parent_pk)}/{child_endpoint}"

        logger.log(
            3,
            f"Adding relation method ${child_endpoint}_to_url$ to parent model -{parent.__name__}- linking to -{child.__name__}-.",
        )
        setattr(parent, f"{child_endpoint.replace('-', '_')}_to_url", to_url)

    def _add_to_created_routes(self, **kwargs: Dict[str, Any]):
        """Add a route to the created routes dictionary.

        Args:
            **kwargs (Dict[str, Any]): Dictionary of keyword arguments.
        """
        model = kwargs.get("child_model", kwargs.get("model"))
        route_key = kwargs["name"]
        self.created_routes[route_key] = {
            "function": route_key,
            "model": model,
            "name": route_key,
            "method": kwargs["method"],
            "url": kwargs["url"],
            "input_schema": kwargs.get("input_schema"),
            "output_schema": kwargs.get("output_schema"),
        }

    def _get_url_naming_function(
        self, model: Callable, input_schema: Callable, output_schema: Callable
    ) -> str:
        """Get the URL naming function for a model.

        Args:
            model (Callable): The model to generate the URL for.
            input_schema (Callable): The input schema class.
            output_schema (Callable): The output schema class.

        Returns:
            str: The URL naming string.
        """
        return get_config_or_model_meta(
            "API_ENDPOINT_NAMER", model, default=endpoint_namer
        )(model, input_schema, output_schema)



import os

import jwt
import datetime

from flask import current_app
from sqlalchemy.exc import NoResultFound

from flarchitect.database.utils import get_primary_keys
from flarchitect.exceptions import CustomHTTPException
from flarchitect.utils.config_helpers import get_config_or_model_meta

# Secret keys (keep them secure)


# In-memory store for refresh tokens (use a persistent database in production)
refresh_tokens_store = {}

def get_pk_and_lookups():
    """
    Get the primary key and lookup field for the user model.
    """
    lookup_field = get_config_or_model_meta("API_USER_LOOKUP_FIELD")
    usr = get_config_or_model_meta("API_USER_MODEL")
    primary_keys = get_primary_keys(usr)
    return primary_keys.name, lookup_field

def generate_access_token(usr_model, expires_in_minutes=360):
    """
    Generates a short-lived access token (JWT).
    """
    pk, lookup_field = get_pk_and_lookups()

    ACCESS_SECRET_KEY = os.environ.get('ACCESS_SECRET_KEY') or current_app.config.get('ACCESS_SECRET_KEY')

    payload = {
        lookup_field: str(getattr(usr_model, lookup_field)),  # Convert UUID to string
        pk: str(getattr(usr_model, pk)),                      # Convert UUID to string
        'exp': datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(minutes=expires_in_minutes),
        'iat': datetime.datetime.now(datetime.timezone.utc)
    }
    token = jwt.encode(payload, ACCESS_SECRET_KEY, algorithm='HS256')
    return token

def generate_refresh_token(usr_model, expires_in_days=2):
    """
    Generates a long-lived refresh token (JWT).
    """
    REFRESH_SECRET_KEY = os.environ.get('REFRESH_SECRET_KEY') or current_app.config.get('REFRESH_SECRET_KEY')

    pk, lookup_field = get_pk_and_lookups()

    payload = {
        lookup_field: str(getattr(usr_model, lookup_field)),  # Convert UUID to string
        pk: str(getattr(usr_model, pk)),                      # Convert UUID to string
        'exp': datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=expires_in_days),
        'iat': datetime.datetime.now(datetime.timezone.utc)
    }
    token = jwt.encode(payload, REFRESH_SECRET_KEY, algorithm='HS256')

    # Store the refresh token in the server-side store
    refresh_tokens_store[token] = {
        lookup_field: str(getattr(usr_model, lookup_field)),  # Convert UUID to string
        pk: str(getattr(usr_model, pk)),                      # Convert UUID to string
        'expires_at': payload['exp']
    }
    return token

def decode_token(token, secret_key):
    """
    Decodes a JWT token.
    """
    try:
        payload = jwt.decode(token, secret_key, algorithms=['HS256'])
        return payload
    except jwt.ExpiredSignatureError:
        raise CustomHTTPException(status_code=401, reason='Token has expired')

    except jwt.InvalidTokenError:
         raise CustomHTTPException(status_code=401, reason='Invalid token')

def refresh_access_token(refresh_token):
    """
    Uses a refresh token to generate a new access token and returns the user.

    Args:
        refresh_token (str): The JWT refresh token.

    Returns:
        tuple: A tuple containing the new access token (str) and the user object.

    Raises:
        CustomHTTPException: If the token is invalid, expired, or the user is not found.
    """
    # Verify refresh token
    REFRESH_SECRET_KEY = os.environ.get('REFRESH_SECRET_KEY') or current_app.config.get('REFRESH_SECRET_KEY')
    payload = decode_token(refresh_token, REFRESH_SECRET_KEY)
    if payload is None:
        raise CustomHTTPException(status_code=401, reason='Invalid token')

    # Check if the refresh token is in the store and not expired
    stored_token = refresh_tokens_store.get(refresh_token)
    if not stored_token or datetime.datetime.now(datetime.timezone.utc) > stored_token['expires_at']:
        raise CustomHTTPException(status_code=403, reason='Invalid or expired refresh token')

    # Get user identifiers from stored_token
    pk_field, lookup_field = get_pk_and_lookups()
    lookup_value = stored_token.get(lookup_field)
    pk_value = stored_token.get(pk_field)

    # Get the user model (this is the SQLAlchemy model)
    usr_model_class = get_config_or_model_meta("API_USER_MODEL")

    # Query the user by lookup_field and pk
    try:
        user = usr_model_class.get_session().query(usr_model_class).filter(
            getattr(usr_model_class, lookup_field) == lookup_value,
            getattr(usr_model_class, pk_field) == pk_value
        ).one()
    except NoResultFound:
        raise CustomHTTPException(status_code=404, reason="User not found")

    # Generate new access token
    new_access_token = generate_access_token(user)

    refresh_tokens_store.pop(refresh_token)

    return new_access_token, user


def get_user_from_token(token, secret_key=None):
    """
    Decodes a JWT token and returns the user associated with the token.

    Args:
        token (str): The JWT token containing user information.
        secret_key (str, optional): The secret key used to decode the token.
                                     If None, uses access secret key.

    Returns:
        usr_model: The user object.

    Raises:
        CustomHTTPException: If the token is invalid or the user is not found.
    """
    # Decode the token
    if secret_key is None:
        ACCESS_SECRET_KEY = os.environ.get('ACCESS_SECRET_KEY') or current_app.config.get('ACCESS_SECRET_KEY')
    else:
        ACCESS_SECRET_KEY = secret_key

    payload = decode_token(token, ACCESS_SECRET_KEY)

    # Get user lookup field and primary key
    pk, lookup_field = get_pk_and_lookups()

    # Get the user model (this is the SQLAlchemy model)
    usr_model_class = get_config_or_model_meta("API_USER_MODEL")

    # Query the user by primary key or lookup field (like username)
    try:
        user = usr_model_class.get_session().query(usr_model_class).filter(
            getattr(usr_model_class, lookup_field) == payload[lookup_field],
            getattr(usr_model_class, pk) == payload[pk]
        ).one()
    except NoResultFound:
        raise CustomHTTPException(status_code=404, reason="User not found")

    return user



import copy
import os
import random
from copy import deepcopy
from datetime import datetime, timedelta
from typing import Dict, Any, Optional, List, Callable, Type

import pytz
from flask import current_app
from marshmallow import Schema, fields
from marshmallow.schema import SchemaMeta
from marshmallow_sqlalchemy.fields import RelatedList, Related, Nested
from sqlalchemy.orm import DeclarativeBase
from werkzeug.http import HTTP_STATUS_CODES

from werkzeug.routing import IntegerConverter, UnicodeConverter

import flarchitect.utils
import flarchitect.utils.decorators
from flarchitect.database.operations import get_model_relationships, get_model_columns
from flarchitect.logging import logger
from flarchitect.utils.general import (
    DATE_FORMAT,
    DATETIME_FORMAT,
    html_path,
    manual_render_absolute_template,
    pluralize_last_word,
)
from flarchitect.utils.core_utils import convert_case
from flarchitect.utils.config_helpers import get_config_or_model_meta
from flarchitect.database.utils import AGGREGATE_FUNCS


def scrape_extra_info_from_spec_data(
    spec_data: Dict[str, Any],
    method: str,
    multiple: bool = False,
    summary: bool = False,
) -> Dict[str, Any]:
    """
    Scrapes extra information from the spec data for API documentation generation.

    Args:
        spec_data (Dict[str, Any]): The spec data.
        method (str): The HTTP method (e.g., 'get', 'post').
        multiple (bool, optional): Whether the operation involves multiple items.
        summary (bool, optional): Whether to generate a summary.

    Returns:
        Dict[str, Any]: Updated spec data with extra information.
    """
    model = spec_data.get("model")
    output_schema = spec_data.get("output_schema")
    input_schema = spec_data.get("input_schema")
    function = spec_data.get("function")

    if not all([model, output_schema or input_schema, method, function]):
        logger.log(1, "Missing data for documentation generation")

    if spec_data.get("tag") is None:
        new_tag = get_config_or_model_meta(
            "tag", model, output_schema, input_schema, "Unknown"
        )
        if new_tag:
            spec_data["tag"] = new_tag
    else:
        spec_data["tag"] = spec_data.get("tag")

    if not summary and get_config_or_model_meta("AUTO_NAME_ENDPOINTS", default=True):
        schema = spec_data.get("output_schema") or spec_data.get("input_schema")
        if schema:
            spec_data["summary"] = make_endpoint_description(
                schema, method, **spec_data
            )
    else:
        spec_data["summary"] = summary

    new_description = get_summary_description(function)
    if new_description:
        spec_data["description"] = new_description

    for description_type in ["summary", "description"]:
        if method.lower() == "get" and multiple:
            config_val = f"{method.lower()}_many" + description_type
        elif method.lower() == "get" and not multiple:
            config_val = f"{method.lower()}_single" + description_type
        else:
            config_val = f"{method.lower()}" + description_type

        new_desc = get_config_or_model_meta(
            config_val, model, output_schema, input_schema, None
        )
        if new_desc:
            spec_data[description_type] = new_desc

    return spec_data


def get_summary_description(f: Callable) -> Optional[str]:
    """
    Retrieves and formats the summary description from a function's docstring.

    Args:
        f (Callable): The function to extract the description from.

    Returns:
        Optional[str]: The formatted description, or None if not available.
    """
    description = f.__doc__.strip() if f.__doc__ else None
    return description.replace("  ", " ") if description else None


def get_param_schema(converter) -> Dict[str, str]:
    """Helper function to get the schema for a parameter based on its converter.

    Args:
        converter: The converter to determine the schema for.

    Returns:
        Dict[str, str]: The schema for the parameter.
    """
    if isinstance(converter, IntegerConverter):
        return {"type": "integer"}
    elif isinstance(converter, UnicodeConverter):
        return {"type": "string"}
    else:
        return {"type": "string"}


# def generate_delete_query_params(schema: Schema, model: DeclarativeBase) -> List[Dict[str, Any]]:
#     """Helper function to generate query parameters for DELETE method.
#
#     Args:
#         schema (Schema): The schema associated with the model.
#         model (DeclarativeBase): The SQLAlchemy model.
#
#     Returns:
#         List[Dict[str, Any]]: List of query parameters for DELETE method.
#     """
#     query_params = []
#     if get_config_or_model_meta("API_ALLOW_DELETE_RELATED", getattr(schema.Meta, "model", None), default=True):
#         related = [x[1] for x in get_model_relationships(model)]
#         query_params.append(
#             {
#                 "name": "delete_related",
#                 "in": "query",
#                 "schema": {"type": "string"},
#                 "description": f"A comma-separated list of related resources to delete. Options include: {', '.join(related)}",
#             }
#         )
#     if get_config_or_model_meta("API_ALLOW_DELETE_DEPENDENTS", getattr(schema.Meta, "model", None), default=True):
#         query_params.append(
#             {
#                 "name": "delete_dependents",
#                 "in": "query",
#                 "schema": {"type": "boolean"},
#                 "description": "If true, will delete all recursively dependent resources.",
#             }
#         )
#     return query_params


def generate_delete_query_params(
    schema: Schema, model: DeclarativeBase
) -> List[Dict[str, Any]]:
    """Helper function to generate query parameters for DELETE method.

    Args:
        schema (Schema): The schema associated with the model.
        model (DeclarativeBase): The SQLAlchemy model.

    Returns:
        List[Dict[str, Any]]: List of query parameters for DELETE method.
    """
    query_params = []

    if get_config_or_model_meta(
        "API_ALLOW_CASCADE_DELETE", getattr(schema.Meta, "model", None), default=True
    ):
        query_params.append(
            {
                "name": "cascade_delete",
                "in": "query",
                "schema": {"type": "boolean"},
                "description": "If true or 1, will delete all recursively dependent resources.",
            }
        )
    return query_params


def generate_get_query_params(
    schema: Schema, model: DeclarativeBase
) -> List[Dict[str, Any]]:
    """Helper function to generate query parameters for GET method.

    Args:
        schema (Schema): The schema associated with the model.
        model (DeclarativeBase): The SQLAlchemy model.

    Returns:
        List[Dict[str, Any]]: List of query parameters for GET method.
    """
    query_params = []
    page_max = get_config_or_model_meta("API_PAGINATION_SIZE_MAX", default=100)
    page_default = get_config_or_model_meta("API_PAGINATION_SIZE_DEFAULT", default=20)

    if get_config_or_model_meta("API_SOFT_DELETE", default=False):
        query_params.append(
            {
                "name": "include_deleted",
                "in": "query",
                "schema": {"type": "boolean"},
                "description": "If true, deleted items will be included in the response.",
            }
        )
    query_params.extend(
        [
            {
                "name": "limit",
                "in": "query",
                "schema": {"type": "integer", "example": 20},
                "description": f"The maximum number of items to return in the response. Default `{page_default}` Maximum `{page_max}`.",
            },
            {
                "name": "page",
                "in": "query",
                "schema": {"type": "integer", "example": 1},
                "description": "The pagination page number. Default `1`.",
            },
        ]
    )
    return query_params


def generate_additional_query_params(
    methods: set, schema: Schema, model: DeclarativeBase
) -> List[Dict[str, Any]]:
    """Helper function to generate additional query parameters.

    Args:
        methods (set): Set of methods to generate query parameters for.
        schema (Schema): The schema associated with the model.
        model (DeclarativeBase): The SQLAlchemy model.

    Returns:
        List[Dict[str, Any]]: List of additional query parameters.
    """
    query_params = []
    for method in methods:
        additional_qs = get_config_or_model_meta(
            "API_ADDITIONAL_QUERY_PARAMS",
            model=model,
            method=method,
            input_schema=schema,
        )
        if additional_qs:
            query_params.extend(additional_qs)
    return query_params


def _add_request_body_to_spec_template(
    spec_template: Dict[str, Any],
    http_method: str,
    input_schema: Schema,
    model: Optional[DeclarativeBase],
):
    """Helper function to add a request body to the spec template.

    Args:
        spec_template (Dict[str, Any]): The OpenAPI specification template to enhance.
        http_method (str): The HTTP method (GET, POST, PUT, DELETE, PATCH).
        input_schema (Schema): The Marshmallow schema for request body validation.
        model (Optional[DeclarativeBase]): The SQLAlchemy model for database interactions.

    Returns:
        None
    """
    case = get_config_or_model_meta("API_SCHEMA_CASE", model=model, default="camel")
    name = convert_case(
        ("patch_" if http_method == "PATCH" else "")
        + input_schema.__name__.replace("Schema", ""),
        case,
    )

    spec_template["requestBody"] = {
        "description": f"`{name}` payload.",
        "required": True,
        "content": {
            "application/json": {
                "schema": {"$ref": f"#/components/schemas/{name}"},
            }
        },
    }


def _add_response_to_spec_template(
    spec_template: Dict[str, Any], output_schema: Schema
):
    """Helper function to add a response to the spec template.

    Args:
        spec_template (Dict[str, Any]): The OpenAPI specification template to enhance.
        output_schema (Schema): The Marshmallow schema for response data serialization.

    Returns:
        None
    """
    model = output_schema.get_model() if hasattr(output_schema, "get_model") else None
    case = get_config_or_model_meta(
        "API_SCHEMA_CASE", model=model, default="camel"
    )
    name = convert_case(output_schema.__name__.replace("Schema", ""), case)

    spec_template.setdefault("responses", {}).setdefault("200", {}).update(
        {
            "description": HTTP_STATUS_CODES.get(200),
            "content": {
                "application/json": {
                    "schema": {"$ref": f"#/components/schemas/{name}"},
                }
            },
        }
    )


def _initialize_base_responses(
    method: str, many: bool, error_responses: List[int]
) -> Dict[str, Dict[str, Any]]:
    """Helper function to initialize base responses.

    Args:
        method (str): The HTTP method.
        many (bool): Whether the endpoint returns multiple items.
        error_responses (List[int]): List of error response status codes.

    Returns:
        Dict[str, Dict[str, Any]]: Dictionary of base responses.
    """
    responses = {"200": {"description": "Successful operation"}}

    if 500 in error_responses or not error_responses:
        responses["500"] = {"description": HTTP_STATUS_CODES.get(500)}

    if (
        method != "POST"
        and not many
        and (404 in error_responses or not error_responses)
    ):
        responses["404"] = {"description": HTTP_STATUS_CODES.get(404)}

    if (
        method == "DELETE"
        and not many
        and (409 in error_responses or not error_responses)
    ):
        responses["409"] = {"description": HTTP_STATUS_CODES.get(409)}

    return responses


def _initialize_auth_responses(error_responses: List[int]) -> Dict[str, Dict[str, Any]]:
    """Helper function to initialize authentication responses.

    Args:
        error_responses (List[int]): List of error response status codes.

    Returns:
        Dict[str, Dict[str, Any]]: Dictionary of authentication responses.
    """
    responses = {}
    auth_on = get_config_or_model_meta("API_AUTHENTICATE", default=False)

    if auth_on and (403 in error_responses or not error_responses):
        responses["403"] = {"description": HTTP_STATUS_CODES.get(403)}

    if auth_on and (401 in error_responses or not error_responses):
        responses["401"] = {"description": HTTP_STATUS_CODES.get(401)}

    return responses


def _initialize_rate_limit_responses(rate_limit: bool) -> Dict[str, Dict[str, Any]]:
    """Helper function to initialize rate limit responses.

    Args:
        rate_limit (bool): Whether the endpoint has a rate limit.

    Returns:
        Dict[str, Dict[str, Any]]: Dictionary of rate limit responses.
    """
    if not rate_limit:
        return {}

    rate_limit_headers = {
        "X-RateLimit-Limit": {
            "description": "The maximum number of requests allowed in a time window.",
            "schema": {"type": "integer", "format": "int32", "example": 100},
        },
        "X-RateLimit-Remaining": {
            "description": "The number of requests remaining in the current rate limit window.",
            "schema": {"type": "integer", "format": "int32", "example": 99},
        },
        "X-RateLimit-Reset": {
            "description": "The time at which the current rate limit window resets (in UTC epoch seconds).",
            "schema": {"type": "integer", "format": "int32", "example": 15830457},
        },
        "Retry-After": {
            "description": "The amount of time to wait before making another request (in seconds).",
            "schema": {"type": "integer", "format": "int32", "example": 20},
        },
    }

    return {
        "429": {
            "description": HTTP_STATUS_CODES.get(429),
            "headers": rate_limit_headers.copy(),
        },
        "200": {
            "headers": rate_limit_headers,
        },
    }


def get_template_data_for_model(schema: Schema) -> Optional[Dict[str, Any]]:
    """Generates model data for Jinja template and Redoc.

    Args:
        schema (Schema): Schema to generate model data for.

    Returns:
        Optional[Dict[str, Any]]: Model data for the template.
    """
    schema_case = get_config_or_model_meta("API_SCHEMA_CASE", default="camel")

    if hasattr(schema, "Meta") and hasattr(schema.Meta, "model"):
        base_model = schema.Meta.model
        base_resource = convert_case(base_model.__name__, schema_case)
        base_fields = get_model_columns(base_model)

        model_relationships = get_model_relationships(base_model)
        model_relationship_names = [
            convert_case(x.__name__, schema_case) for x in model_relationships
        ]

        if model_relationships:
            relationship_fields = get_model_columns(model_relationships[0])
            relationship_resource = convert_case(
                model_relationships[0].__name__, schema_case
            )
        else:
            relationship_fields = []
            relationship_resource = None

        aggs = ", ".join([f"`{x}`" for x in AGGREGATE_FUNCS.keys()])

        return {
            "relationship_resource": relationship_resource,
            "relationship_fields": relationship_fields,
            "base_resource": base_resource,
            "base_fields": base_fields,
            "aggs": aggs,
            "model_relationship_names": model_relationship_names,
        }
    return None


def generate_example_values(
    now: datetime, yesterday: datetime, day_before_yesterday: datetime
) -> Dict[str, List[str]]:
    """Helper function to generate example values for filter examples.

    Args:
        now (datetime): Current datetime.
        yesterday (datetime): Datetime of yesterday.
        day_before_yesterday (datetime): Datetime of the day before yesterday.

    Returns:
        Dict[str, List[str]]: Dictionary with example values.
    """
    return {
        "Integer": ["1", "10", "100", "500", "1000"],
        "Float": ["1.25", "2.50", "3.75", "5.00"],
        "String": ["John", "Doe", "Jane"],
        "Boolean": ["true", "false"],
        "Date": [
            now.date().strftime(DATE_FORMAT),
            yesterday.date().strftime(DATE_FORMAT),
            day_before_yesterday.date().strftime(DATE_FORMAT),
        ],
        "DateTime": [
            now.strftime(DATETIME_FORMAT),
            yesterday.strftime(DATETIME_FORMAT),
            day_before_yesterday.strftime(DATETIME_FORMAT),
        ],
        "Time": ["12:00:00", "13:00:00", "14:00:00"],
    }


def generate_operators() -> Dict[str, List[str]]:
    """Helper function to generate operators for different data types.

    Returns:
        Dict[str, List[str]]: Dictionary with operators for each data type.
    """
    return {
        "Integer": [
            "__eq",
            "__lt",
            "__le",
            "__gt",
            "__ge",
            "__ne",
            "__in",
            "__nin",
            "__like",
            "__ilike",
        ],
        "Float": [
            "__eq",
            "__lt",
            "__le",
            "__gt",
            "__ge",
            "__ne",
            "__in",
            "__nin",
            "__like",
            "__ilike",
        ],
        "String": ["__eq", "__ne", "__in", "__nin", "__like", "__ilike"],
        "Bool": ["__eq", "__ne", "__in", "__nin"],
        "Date": ["__eq", "__lt", "__le", "__gt", "__ge", "__ne", "__in", "__nin"],
        "DateTime": ["__eq", "__lt", "__le", "__gt", "__ge", "__ne", "__in", "__nin"],
        "Time": ["__eq", "__lt", "__le", "__gt", "__ge", "__ne", "__in", "__nin"],
    }


def generate_operator_examples(
    schema: Schema,
    operators: Dict[str, List[str]],
    example_values: Dict[str, List[str]],
) -> List[str]:
    """Helper function to generate operator examples for filters.

    Args:
        schema (Schema): The schema to generate examples for.
        operators (Dict[str, List[str]]): Operators applicable to each data type.
        example_values (Dict[str, List[str]]): Example values for each data type.

    Returns:
        List[str]: List of generated filter examples.
    """
    examples = []

    # Assuming fields is a function that needs to be called
    try:
        fields = schema().dump_fields  # Accessing fields through an instance
    except TypeError:
        fields = schema.dump_fields


    # Generate column examples only for those fields that are not dump_only
    columns = [k for k, v in fields.items() if v and not v.dump_only]

    for column in columns:
        col_type = type(fields[column]).__name__
        if col_type in operators:
            chosen_operator = random.choice(operators[col_type])
            if chosen_operator in ["__in", "__nin"]:
                chosen_values = ", ".join(
                    random.choices(example_values.get(col_type, ["value"]), k=3)
                )
                examples.append(f"{column}{chosen_operator}=({chosen_values})")
            else:
                chosen_value = random.choice(example_values.get(col_type, ["value"]))
                examples.append(f"{column}{chosen_operator}={chosen_value}")

    return examples


def get_table_name(
    index: int, resource_name: str, fields: List[str], example_table: List[str]
) -> str:
    """Helper function to generate table names for field examples.

    Args:
        index (int): The index to determine which example to generate.
        resource_name (str): The name of the resource.
        fields (List[str]): The list of fields.
        example_table (List[str]): The example table data.

    Returns:
        str: Generated table name.
    """
    temp_fields = copy.deepcopy([x[0] for x in fields])
    random.shuffle(temp_fields)
    if index == 0:
        current_fields = temp_fields[:3]
    elif index == 1:
        current_fields = [resource_name + "." + x for x in temp_fields[:3]]
    else:
        current_fields = []
        for _ in range(5):
            table_choice = random.choice([resource_name, "OtherTable"])
            field_choice = (
                random.choice(temp_fields)
                if table_choice == resource_name
                else random.choice(example_table)
            )
            current_fields.append(f"{table_choice}.{field_choice}")

    return "fields=" + ",".join(current_fields)


def _prepare_patch_schema(input_schema: Optional[Schema]) -> Optional[Schema]:
    """Prepare a patch schema with all fields optional.

    Args:
        input_schema (Optional[Schema]): The input schema to prepare the patch schema for.

    Returns:
        Optional[Schema]: The prepared patch schema with all fields set to optional.
    """
    if not input_schema:
        return None

    # Initialize a dictionary to hold the new class's fields
    class_fields = {}

    # Iterate over all fields in the input schema
    items = input_schema.fields.items() if hasattr(input_schema, "fields") else input_schema().fields.items()

    for field_name, field_obj in items:
        # Deepcopy the field to avoid mutating the original field
        new_field = deepcopy(field_obj)

        # Set the field to be optional
        new_field.required = False
        new_field.allow_none = True

        # Add the modified field to the class fields
        class_fields[field_name] = new_field

    # Preserve the Meta class if it exists
    Meta = getattr(input_schema.__class__, 'Meta', None)
    if Meta:
        class_fields['Meta'] = Meta

    # Dynamically create a new schema class with the modified fields
    PatchSchemaClass = type(
        f"Patch{input_schema.__class__.__name__}",
        (Schema,),  # Inherit directly from Schema to avoid field conflicts
        class_fields
    )

    # Instantiate the new schema class
    put_input_schema = PatchSchemaClass()

    return put_input_schema

def make_endpoint_description(schema: Schema, http_method: str, **kwargs) -> str:
    """Generates endpoint description from a schema for the API docs.

    Only applicable in FULL_AUTO mode or if AUTO_NAME_ENDPOINTS = True.

    Args:
        schema (Schema): Schema to generate endpoint description from.
        http_method (str): HTTP method.

    Returns:
        str: Endpoint description.
    """
    many = kwargs.get("multiple")
    model = getattr(schema, "get_model", lambda: None)()
    name = (kwargs.get("model") or model or schema).__name__.replace("Schema", "")
    name = convert_case(
        name, get_config_or_model_meta("API_SCHEMA_CASE", model=model, default="camel")
    )

    parent = kwargs.get("parent")
    parent_name = parent.__name__ if parent else ""

    if http_method == "GET":
        if parent and many:
            return f"Returns a list of `{name}` for a specific `{parent_name}`"
        elif parent and not many:
            return f"Get a `{name}` by id for a specific `{parent_name}`."
        elif many:
            return f"Returns a list of `{name}`"
        else:
            return f"Get a `{name}` by id."
    elif http_method == "POST":
        return f"Create a new `{name}`."
    elif http_method in ["PUT", "PATCH"]:
        return f"Update an existing `{name}`."
    elif http_method == "DELETE":
        return f"Delete a `{name}` by id."
    else:
        return "Endpoint description not available"


def generate_fields_description(schema: Schema) -> str:
    """Generates fields description from a schema for the API docs.

    Args:
        schema (Schema): Schema to generate fields description from.

    Returns:
        str: Fields description.
    """

    if callable(schema):
        schema = schema()

    fields = [
        (k, v.metadata.get("description", ""))
        for k, v in schema.fields.items()
        if v and v.dump_only is False and not isinstance(v, (RelatedList, Related))
    ]

    if hasattr(schema, "Meta") and hasattr(schema.Meta, "model"):
        resource_name = schema.Meta.model.__name__
        example_table = (
            "OtherTable.name,OtherTable.age,OtherTable.id,OtherTable.email".split(",")
        )
        example_fields = [
            get_table_name(i, resource_name, fields, example_table) for i in range(3)
        ]

        full_path = os.path.join(html_path, "redoc_templates/fields.html")
        schema_name = endpoint_namer(schema.Meta.model)
        api_prefix = get_config_or_model_meta("API_PREFIX", default="/api")

        return manual_render_absolute_template(
            full_path,
            schema_name=schema_name,
            api_prefix=api_prefix,
            fields=fields,
            example_fields=example_fields,
        )

    return "None"


def generate_x_description(template_data: Dict, path: str = "") -> str:
    """Generates filter examples from a model.

    Args:
        template_data (dict): Template data to generate filter examples from.
        path (str): Path to the template.

    Returns:
        str: Filter examples.
    """
    if template_data:
        full_path = os.path.join(html_path, path)
        return manual_render_absolute_template(full_path, **template_data)
    else:
        return "This endpoint does not have a database table (or is computed etc) and should not be filtered\n"


def generate_filter_examples(schema: Schema) -> str:
    """Generates filter examples from a model.

    Args:
        schema (Schema): Schema to generate filter examples from.

    Returns:
        str: Filter examples.
    """
    now = datetime.now(pytz.utc)
    yesterday = now - timedelta(days=1)
    day_before_yesterday = yesterday - timedelta(days=1)

    example_values = generate_example_values(now, yesterday, day_before_yesterday)
    operators = generate_operators()

    examples = generate_operator_examples(schema, operators, example_values)

    split_examples = len(examples) // 3
    example_one = "&".join(examples[:split_examples])
    example_two = "&".join(examples[-split_examples:])

    full_path = os.path.join(html_path, "redoc_templates/filters.html")

    return manual_render_absolute_template(
        full_path, examples=[example_one, example_two]
    )


def convert_path_to_openapi(path: str) -> str:
    """Converts a Flask path to an OpenAPI path.

    Args:
        path (str): Flask path to convert.

    Returns:
        str: OpenAPI path.
    """
    return path.replace("<", "{").replace(">", "}").replace("<int:", "")


def initialize_spec_template(
    method: str,
    many: bool = False,
    rate_limit: bool = False,
    error_responses: Optional[List[int]] = None,
) -> Dict[str, Any]:
    """Initializes the spec template with optional rate limiting headers for successful and rate-limited responses.

    Args:
        method (str): The HTTP method.
        many (bool): Whether the endpoint returns multiple items.
        rate_limit (bool): Whether the endpoint has a rate limit.
        error_responses (Optional[List[int]]): List of error response status codes.

    Returns:
        Dict[str, Any]: Spec template.
    """
    if not error_responses:
        error_responses = []

    responses = _initialize_base_responses(method, many, error_responses)
    responses.update(_initialize_auth_responses(error_responses))
    responses.update(_initialize_rate_limit_responses(rate_limit))

    return {"responses": responses, "parameters": []}


def append_parameters(
    spec_template: Dict[str, Any],
    query_params: List[Dict[str, Any]],
    path_params: List[Dict[str, Any]],
    http_method: str,
    input_schema: Optional[Schema] = None,
    output_schema: Optional[Schema] = None,
    model: Optional[DeclarativeBase] = None,
    many: bool = False,
):
    """Enhances a spec template with parameters, request bodies, and responses for API documentation.

    Args:
        spec_template (Dict[str, Any]): The OpenAPI specification template to enhance.
        query_params (List[Dict[str, Any]]): A list of dictionaries defining query parameters.
        path_params (List[Dict[str, Any]]): A list of dictionaries defining path parameters.
        http_method (str): The HTTP method (GET, POST, PUT, DELETE, PATCH).
        input_schema (Optional[Schema]): The Marshmallow schema for request body validation.
        output_schema (Optional[Schema]): The Marshmallow schema for response data serialization.
        model (Optional[DeclarativeBase]): The SQLAlchemy model for database interactions.
        many (bool): Whether the endpoint returns multiple items.

    Returns:
        None
    """
    flarchitect.utils.general.html_path = current_app.extensions[
        "flarchitect"
    ].get_templates_path()

    spec_template.setdefault("parameters", []).extend(path_params + query_params)
    rate_limit = get_config_or_model_meta(
        "API_RATE_LIMIT",
        model=model,
        input_schema=input_schema,
        output_schema=output_schema,
        default=False,
    )

    if rate_limit:
        description = spec_template.get("description", "")
        spec_template["description"] = (
            f"{description}\n**Rate Limited** - requests on this endpoint are limited to `{rate_limit}`."
        )

    if input_schema:
        _add_request_body_to_spec_template(
            spec_template, http_method, input_schema, model
        )

    if output_schema:
        _add_response_to_spec_template(spec_template, output_schema)

    if (
        http_method == "GET"
        and model
        and many
        and get_config_or_model_meta("API_ALLOW_FILTERS", model=model, default=True)
    ):
        spec_template["parameters"].append(
            {
                "name": "filters",
                "in": "query",
                "schema": {"type": "string"},
                "description": generate_filter_examples(output_schema),
            }
        )

        template_data = get_template_data_for_model(output_schema)
        spec_template["parameters"].extend(
            make_endpoint_params_description(output_schema, template_data)
        )

    add_auth_to_spec(model, spec_template)


def add_auth_to_spec(model: DeclarativeBase, spec_template: Dict[str, Any]):
    """Adds authentication information to the spec template.

    Args:
        model (DeclarativeBase): The SQLAlchemy model.
        spec_template (Dict[str, Any]): The OpenAPI specification template to enhance.

    Returns:
        None
    """
    auth_on = get_config_or_model_meta("API_AUTHENTICATE", model=model, default=False)
    auth_type = get_config_or_model_meta(
        "API_AUTHENTICATE_METHOD", model=model, default=None
    )

    if not auth_on:
        return

    spec_template.setdefault("security", [])
    spec_template.setdefault("components", {}).setdefault("securitySchemes", {})

    if auth_type == "basic":
        spec_template["security"].append({"basicAuth": []})
        spec_template["components"]["securitySchemes"]["basicAuth"] = {
            "type": "http",
            "scheme": "basic",
            "description": "Basic Authentication. Credentials must be provided as a Base64-encoded string in the format `username:password` in the `Authorization` header.",
        }
    elif auth_type == "jwt":
        spec_template["security"].append({"bearerAuth": []})
        spec_template["components"]["securitySchemes"]["bearerAuth"] = {
            "type": "http",
            "scheme": "bearer",
            "bearerFormat": "JWT",
            "description": "JWT Authentication. A valid JWT must be sent in the `Authorization` header as `Bearer <token>`.",
        }
    elif auth_type == "api_key":
        api_key_location = "header"
        api_key_name = "X-API-KEY"
        spec_template["security"].append({"apiKeyAuth": []})
        spec_template["components"]["securitySchemes"]["apiKeyAuth"] = {
            "type": "apiKey",
            "in": api_key_location,
            "name": api_key_name,
            "description": f"API Key Authentication. An API key must be sent in the `{api_key_name}` {api_key_location}.",
        }


def make_endpoint_params_description(
    schema: Schema, data: Dict[str, Any]
) -> List[Dict[str, Any]]:
    """Generates endpoint parameters description from a schema for the API docs.

    Args:
        schema (Schema): Schema to generate endpoint parameters description from.
        data (Dict[str, Any]): Data to generate endpoint parameters description from.

    Returns:
        List[Dict[str, Any]]: Endpoint parameters description.
    """
    params = []

    if get_config_or_model_meta(
        "API_ALLOW_SELECT_FIELDS", getattr(schema.Meta, "model", None), default=True
    ):
        params.append(
            {
                "name": "fields",
                "in": "query",
                "schema": {"type": "string"},
                "description": generate_fields_description(schema),
            }
        )

    if get_config_or_model_meta(
        "API_ALLOW_ORDER_BY", getattr(schema.Meta, "model", None), default=True
    ):
        params.append(
            {
                "name": "order by",
                "in": "query",
                "schema": {"type": "string"},
                "description": generate_x_description(
                    data, "redoc_templates/order.html"
                ),
            }
        )

    if get_config_or_model_meta(
        "API_ALLOW_JOIN", getattr(schema.Meta, "model", None), default=False
    ):
        params.append(
            {
                "name": "joins",
                "in": "query",
                "schema": {"type": "string"},
                "description": generate_x_description(
                    data, "redoc_templates/joins.html"
                ),
            }
        )

    if get_config_or_model_meta(
        "API_ALLOW_GROUPBY", getattr(schema.Meta, "model", None), default=False
    ):
        params.append(
            {
                "name": "group by",
                "in": "query",
                "schema": {"type": "string"},
                "description": generate_x_description(
                    data, "redoc_templates/group.html"
                ),
            }
        )

    if get_config_or_model_meta(
        "API_ALLOW_AGGREGATION", getattr(schema.Meta, "model", None), default=False
    ):
        params.append(
            {
                "name": "aggregation",
                "in": "query",
                "schema": {"type": "string"},
                "description": generate_x_description(
                    data, "redoc_templates/aggregate.html"
                ),
            }
        )

    return params


def handle_authorization(f: Callable, spec_template: Dict[str, Any]):
    """Handles authorization in the documentation.

    Args:
        f (Callable): The function to handle authorization for.
        spec_template (Dict[str, Any]): The OpenAPI specification template to enhance.

    Returns:
        None
    """
    roles_required = None

    if hasattr(f, "_decorators"):
        for decorator in f._decorators:
            if decorator.__name__ == "auth_required":
                roles_required = decorator._args
                spec_template["parameters"].append(
                    {
                        "name": "Authorization",
                        "in": "header",
                        "description": "JWT token required",
                        "required": True,
                        "schema": {"type": "string", "format": "jwt"},
                    }
                )
                break

    if roles_required:
        roles_desc = ", ".join(roles_required)
        spec_template["responses"]["401"][
            "description"
        ] += f" Roles required: {roles_desc}."


def get_openapi_meta_data(field_obj: fields.Field) -> Dict[str, Any]:
    """Convert a Marshmallow field to an OpenAPI type.

    Args:
        field_obj (fields.Field): The Marshmallow field to convert.

    Returns:
        Dict[str, Any]: The OpenAPI metadata for the field.
    """
    openapi_type_info = {}
    field_type = type(field_obj)

    if (
        hasattr(field_obj, "parent")
        and hasattr(field_obj.parent, "Meta")
        and hasattr(field_obj.parent.Meta, "model")
    ):
        openapi_type_info = get_description_and_example_add(
            openapi_type_info, field_obj
        )

    openapi_type_info["type"] = type_mapping.get(field_type, "string")

    if field_type in [fields.DateTime, fields.Date, fields.Time]:
        openapi_type_info["format"] = (
            "date-time"
            if field_type == fields.DateTime
            else field_type.__name__.lower()
        )

    if field_type == fields.Function:
        openapi_type_info["format"] = "uri"
        openapi_type_info["example"] = "/url/to/resource"

    if field_type == fields.List:
        openapi_type_info["items"] = get_openapi_meta_data(field_obj.inner)

    if field_type in [Nested, Related, RelatedList]:
        related_schema_name = get_related_schema_name(field_obj, field_type)
        if related_schema_name:
            openapi_type_info = handle_nested_related_fields(
                field_obj, field_type, related_schema_name, openapi_type_info
            )

    return openapi_type_info


def get_related_schema_name(field_obj: fields.Field, field_type: type) -> Optional[str]:
    """Get the related schema name for nested or related fields.

    Args:
        field_obj (fields.Field): The field to inspect.
        field_type (type): The type of the field.

    Returns:
        Optional[str]: The related schema name, if found.
    """
    if field_type == Nested:
        return field_obj.schema.__class__.__name__
    elif field_type in [Related, RelatedList]:
        parent_model = field_obj.parent.Meta.model
        related_model = getattr(parent_model, field_obj.name).property.mapper.class_
        return f"{related_model.__name__}Schema"
    return None


def handle_nested_related_fields(
    field_obj: fields.Field,
    field_type: type,
    related_schema_name: str,
    openapi_type_info: Dict[str, Any],
) -> Dict[str, Any]:
    """Handle the OpenAPI metadata for nested or related fields.

    Args:
        field_obj (fields.Field): The field to handle.
        field_type (type): The type of the field.
        related_schema_name (str): The name of the related schema.
        openapi_type_info (Dict[str, Any]): The OpenAPI metadata to update.

    Returns:
        Dict[str, Any]: The updated OpenAPI metadata.
    """
    if field_obj.many or field_type == RelatedList:
        openapi_type_info["type"] = "array"
        openapi_type_info["items"] = {
            "$ref": f"#/components/schemas/{related_schema_name}"
        }
    else:
        openapi_type_info["$ref"] = f"#/components/schemas/{related_schema_name}"
    return openapi_type_info


def get_description_and_example_add(
    openapi_type_info: Dict[str, Any], field_obj: fields.Field
) -> Dict[str, Any]:
    """Add description and example to the OpenAPI metadata from the model field.

    Args:
        openapi_type_info (Dict[str, Any]): The OpenAPI metadata to update.
        field_obj (fields.Field): The Marshmallow field to inspect.

    Returns:
        Dict[str, Any]: The updated OpenAPI metadata.
    """
    model_field = getattr(field_obj.parent.Meta.model, field_obj.name, None)
    if model_field and hasattr(model_field, "info"):
        model_field_metadata = model_field.info
        if description := model_field_metadata.get("description"):
            openapi_type_info["description"] = description
        if example := model_field_metadata.get("example"):
            openapi_type_info["example"] = example
    return openapi_type_info


type_mapping = {
    fields.String: "string",
    fields.Integer: "integer",
    fields.Float: "number",
    fields.Boolean: "boolean",
    fields.DateTime: "string",
    fields.Date: "string",
    fields.Time: "string",
    fields.UUID: "string",
    fields.URL: "string",
    fields.Function: "string",
    fields.Nested: "object",
    fields.Email: "string",
    fields.Dict: "object",
    fields.List: "array",
    Related: "object",
    RelatedList: "array",
}


def get_description(kwargs: Dict[str, Any]) -> str:
    """
    Get the description for the route from the model's Meta class or a default description based on the method.

    Args:
        kwargs: Keyword arguments including the model and the method type (GET, POST, etc.).

    Returns:
        str: The description string for the route.
    """
    model = kwargs.get("model", kwargs.get("child_model"))
    name, method = kwargs["name"], kwargs["method"]

    if "child_model" in kwargs:
        parent = kwargs["parent_model"]
        return f"Get multiple `{name}` records from the database based on the parent {endpoint_namer(parent)} id"

    description = (
        getattr(model.Meta, "description", {}).get(method)
        if hasattr(model, "Meta")
        else None
    )
    if description:
        return description

    default_descriptions = {
        "DELETE": f"Delete a single `{name}` in the database by its id",
        "PATCH": f"Patch (update) a single `{name}` in the database.",
        "POST": f"Post (create) a single `{name}` in the database.",
        "GET": f"Get {'multiple' if kwargs.get('multiple', False) else 'a single'} `{name}` record(s) from the database",
    }
    return default_descriptions.get(method, "")


def get_tag_group(kwargs: Dict[str, Any]) -> Optional[str]:
    """
    Get the x-tagGroup for the route from the model's Meta class.

    Args:
        kwargs: Keyword arguments including the model and the method type (GET, POST, etc.).

    Returns:
        str: The tag group string for the route.
    """
    model = kwargs.get("model", kwargs.get("child_model"))
    return getattr(model.Meta, "tag_group", None) if hasattr(model, "Meta") else None


def endpoint_namer(
    model: Optional[Type[DeclarativeBase]] = None,
    input_schema: Optional[Type[Schema]] = None,
    output_schema: Optional[Type[Schema]] = None,
) -> str:
    """
    Get the endpoint name for the model based on its name.

    Args:
        model (Optional[Type[DeclarativeBase]]): The model to get the endpoint name for.
        input_schema (Optional[Type[Schema]]): The input schema for the model.
        output_schema (Optional[Type[Schema]]): The output schema for the model.

    Returns:
        str: The endpoint name.
    """
    case = get_config_or_model_meta("API_ENDPOINT_CASE", default="kebab", model=model)
    converted_name = convert_case(model.__name__, case)
    return pluralize_last_word(converted_name)



from setuptools import setup, find_packages

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup(
    name="flarchitect",
    version="0.0.8.24",  # Note the normalized version format
    packages=find_packages(include=[
        'flarchitect',
        'flarchitect.*',
        'flarchitect.html',
        'flarchitect.html.redoc_templates'
    ]),
    license="MIT",
    author="arched.dev (Lewis Morris)",
    author_email="hello@arched.dev",
    description="Automatic RESTful API generator with redoc",
    install_requires=requirements,
    package_data={
        "flarchitect": [
            "html/*",
            "html/redoc_templates/*",
        ],
    },
    include_package_data=True,
)



import datetime
import uuid
from functools import partial
from typing import Any, Optional

from flask import request
from marshmallow import fields, Schema, post_dump, pre_dump
from marshmallow.validate import Length, Range
from sqlalchemy import (
    Integer,
    String,
    Boolean,
    Float,
    Date,
    DateTime,
    Time,
    Text,
    Numeric,
    BigInteger,
    LargeBinary,
    Enum,
    ARRAY,
    Interval,
    Column, inspect, SmallInteger, TIMESTAMP, UUID,
)
from sqlalchemy.dialects.postgresql import JSON, JSONB
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import class_mapper, RelationshipProperty, ColumnProperty, InstrumentedAttribute

from flarchitect.database.utils import _get_relation_use_list_and_type
from flarchitect.logging import logger
from flarchitect.schemas.validators import validate_by_type
from flarchitect.specs.utils import get_openapi_meta_data, endpoint_namer
from flarchitect.schemas.utils import get_input_output_from_model_or_make
from flarchitect.utils.core_utils import convert_case
from flarchitect.utils.config_helpers import get_config_or_model_meta

# Mapping between SQLAlchemy types and Marshmallow fields
type_mapping = {
    Integer: fields.Int,
    SmallInteger: fields.Int,          # Added SmallInteger
    String: fields.Str,
    Text: fields.Str,
    Boolean: fields.Bool,
    Float: fields.Float,
    Date: fields.Date,
    DateTime: fields.DateTime,
    Time: fields.Time,
    TIMESTAMP: fields.DateTime,        # Added TIMESTAMP
    JSON: fields.Raw,
    JSONB: fields.Raw,
    Numeric: fields.Decimal,
    BigInteger: fields.Int,
    LargeBinary: fields.Str,            # Consider using fields.Raw for binary data
    Enum: fields.Str,                   # Consider fields.Enum for stricter validation
    Interval: fields.TimeDelta,
    UUID: fields.UUID,                  # Added UUID
    str: fields.Str,
    int: fields.Int,
    bool: fields.Bool,
    float: fields.Float,
    dict: fields.Dict,
    list: fields.List,
}


class Base(Schema):  # Inheriting from marshmallow's Schema
    @classmethod
    def get_model(cls):
        """Get the SQLAlchemy model associated with the schema."""
        meta = getattr(cls, "Meta", None)
        return getattr(meta, "model", None)


class DeleteSchema(Base):
    complete = fields.Boolean(required=True, default=False)


class AutoSchema(Base):
    class Meta:
        model = None
        add_hybrid_properties = True
        include_children = True

    def __init__(self, *args, render_nested=True, **kwargs):
        """Initialize the AutoSchema instance."""
        self.render_nested = render_nested
        self.depth = kwargs.pop("depth", 0)
        only_fields = kwargs.pop("only", None)

        # Ensure context is set up properly
        if 'context' not in kwargs or kwargs['context'] is None:
            kwargs['context'] = {}

        kwargs['context'].setdefault('current_depth', self.depth)

        super().__init__(*args, **kwargs)

        self.model = self.Meta.model

        if self.model:
            schema_case = get_config_or_model_meta(
                "API_SCHEMA_CASE", model=self.model, default="camel"
            )
            self.__name__ = convert_case(self.model.__name__, schema_case)
            self.generate_fields()

        if only_fields:
            self._apply_only(only_fields)

    @pre_dump
    def pre_dump(self, data, **kwargs):
        # print("Pre-dump data:", data)
        return data  # Ensure the data is returned unchanged

    @post_dump
    def post_dump(self, data: dict, **kwargs) -> dict:
        """Apply a post-dump callback if configured."""
        post_dump_function = get_config_or_model_meta(
            "API_POST_DUMP_CALLBACK",
            model=self.get_model(),
            method=request.method,
            default=None,
        )
        return post_dump_function(data, **kwargs) if post_dump_function else data

    def _apply_only(self, only_fields: list):
        """Filter fields to include only those specified."""
        self.fields = {key: self.fields[key] for key in only_fields}
        # todo Add a check to see if ok to dump or not

        self.dump_fields = {key: self.dump_fields[key] for key in only_fields}
        self.load_fields = {key: self.load_fields[key] for key in only_fields}

    def generate_fields(self):
        """Automatically add fields for each column and relationship in the SQLAlchemy model."""
        if not self.model:
            logger.warning("self.Meta.model is None. Skipping field generation.")
            return

        mapper = class_mapper(self.model)
        for attribute, mapper_property in mapper.all_orm_descriptors.items():
            original_attribute = attribute
            attribute = self._convert_case(attribute)

            if self._should_skip_attribute(attribute):
                continue

            # Access the actual property from the InstrumentedAttribute
            prop = getattr(mapper_property, 'property', None)

            if isinstance(prop, RelationshipProperty):
                self._handle_relationship(
                    attribute, original_attribute, mapper_property
                )
            elif isinstance(prop, ColumnProperty):
                self._handle_column(attribute, original_attribute, mapper_property)

            elif isinstance(mapper_property, hybrid_property):
                if get_config_or_model_meta(
                    "API_DUMP_HYBRID_PROPERTIES", model=self.model, default=True
                ):
                    self._handle_hybrid_property(
                        attribute, original_attribute, mapper_property
                    )
            else:
                pass
        # print("Final fields:", self.fields)
        # print("Dump fields:", self.dump_fields)  # Should match self.fields
        # print("Load fields:", self.load_fields)  # Should match self.fields

    def _convert_case(self, attribute: str) -> str:
        """Convert the attribute name to the appropriate case."""
        field_case = get_config_or_model_meta(
            "API_FIELD_CASE", model=self.model, default="snake_case"
        )
        return convert_case(attribute, field_case)

    def _should_skip_attribute(self, attribute: str) -> bool:
        """Determine if the attribute should be skipped."""
        return attribute.startswith("_") and get_config_or_model_meta(
            "API_IGNORE_UNDERSCORE_ATTRIBUTES", model=self.model, default=True
        )

    def _handle_relationship(
        self,
        attribute: str,
        original_attribute: str,
        mapper_property: RelationshipProperty,
    ):
        """Handle adding a relationship field to the schema."""
        try:
            if request.args.get("dump_relationships") in ["false", "False", "0"]:
                return
        except RuntimeError:
            # this happens when the model is outside the flask request and is expected when the schema is first
            # initialized
            pass

        self.add_relationship_field(attribute, original_attribute, mapper_property)

    def _handle_column(
        self, attribute: str, original_attribute: str, mapper_property: ColumnProperty
    ):
        """Handle adding a column field to the schema."""
        column_type = mapper_property.property.columns[0].type
        self.add_column_field(attribute, original_attribute, column_type)

    def _handle_hybrid_property(
        self, attribute: str, original_attribute: str, mapper_property: hybrid_property
    ):
        """Handle adding a hybrid property field to the schema."""
        self.add_hybrid_property_field(
            attribute,
            original_attribute,
            mapper_property.__annotations__.get("return"),
        )

    def add_hybrid_property_field(
        self, attribute: str, original_attribute: str, field_type: Optional[Any]
    ):
        """Automatically add a field for a given hybrid property in the SQLAlchemy model."""
        if self._should_skip_attribute(attribute):
            return

        field_type = (
            type_mapping.get(field_type, fields.Str) if field_type else fields.Str
        )

        # Check if the attribute has a setter method
        has_setter = (
            hasattr(type(self.model), original_attribute)
            and isinstance(getattr(type(self.model), original_attribute), property)
            and getattr(type(self.model), original_attribute).fset is not None
        )

        # If there's no setter, mark it as dump_only
        field_args = {"dump_only": not has_setter}

        self.add_to_fields(
            original_attribute, field_type(data_key=attribute, **field_args), load=False
        )

        self._update_field_metadata(original_attribute)

    def add_column_field(
        self, attribute: str, original_attribute: str, column_type: Any
    ):
        """Automatically add a field for a given column in the SQLAlchemy model."""

        # Check if the attribute should be skipped
        if self._should_skip_attribute(attribute):
            return

        # Map the SQLAlchemy column type to a Marshmallow field type
        field_type = type_mapping.get(type(column_type))
        if not field_type:
            return

        # Get additional attrs for the field based on the column's properties
        field_args = self._get_column_field_attrs(original_attribute, column_type)

        # Add the field to the schema, using the original attribute name
        self.add_to_fields(
            original_attribute, field_type(data_key=attribute, **field_args)
        )

        # Update the OpenAPI metadata for the field
        self._update_field_metadata(original_attribute)

    def add_to_fields(self, attribute, field, load=True, dump=True):
        """
        Add a field to the fields dictionary
        Args:
            attribute (str): The attribute name
            field (Field): The field object

        Returns:
            None

        """
        self.declared_fields[attribute] = field

        self.fields[attribute] = field
        if load:
            self.load_fields[attribute] = field
        if dump:
            self.dump_fields[attribute] = field

    def _get_column_field_attrs(
        self, original_attribute: str, column_type: Any
    ) -> dict:
        """Get additional arguments for column fields."""
        column = self.model.__table__.columns.get(original_attribute)

        # Check if column is None
        if column is None:
            return {}

        field_args = {}

        # Check for non-nullable columns that are not primary keys and auto-increment
        if (
            not column.nullable
            and not column.primary_key
            and column.autoincrement
            and column.default is None
        ):
            field_args["required"] = True

        # Handle default values for the column
        if column.default:
            field_args["default"] = (
                column.default.arg if not callable(column.default.arg) else None
            )

        field_args["validate"] = []
        # Check for column length constraints and add validation
        field_args = self._add_validation(column, field_args)

        # Mark fields as unique or primary keys
        if column.unique or column.primary_key:
            field_args["unique"] = True

        return field_args

    def _add_validation(self, column: Column, field_args: dict):

        # custom validation by user
        if column.info.get("validate"):
            validator = validate_by_type(column.info.get("validate"))
            if not validator:
                raise ValueError(
                    f"Invalid validator type: model {self.model.__name__}.{column.name} - {column.info.get('validate')}"
                )
            field_args["validate"].append(validator)
            return field_args

        # Add validation to the field based on the column type in sql
        if hasattr(column.type, "length"):
            field_args["validate"].append(Length(max=column.type.length))

        if isinstance(column.type, (Float, Numeric)):
            field_args["validate"].append(Range(min=float("-inf"), max=float("inf")))

        if isinstance(column.type, Integer):
            field_args["validate"].append(Range(min=-2147483648, max=2147483647))

        # auto
        if get_config_or_model_meta(
            "API_AUTO_VALIDATE", model=self.model, default=True
        ):
            # todo add more validation and test
            column_name = column.name
            format_name = column.info.get("format")

            if "email" in column_name or format_name == "email":
                field_args["validate"].append(validate_by_type("email"))
            elif "url" in column_name or format_name in ["url", "uri", "url_path"]:
                field_args["validate"].append(validate_by_type("url"))
            elif (
                "date" in column_name
                or column.type.python_type == datetime.date
                or format_name == "date"
            ):
                field_args["validate"].append(validate_by_type("date"))
            elif (
                column.type.python_type == datetime.time
                or format_name == "time"
            ):
                field_args["validate"].append(validate_by_type("time"))
            elif (
                "datetime" in column_name
                or column.type.python_type == datetime.datetime
                or format_name == "datetime"
            ):
                field_args["validate"].append(validate_by_type("datetime"))
            elif (
                "boolean" in column_name
                or column.type.python_type == bool
                or format_name == "boolean"
            ):
                field_args["validate"].append(validate_by_type("boolean"))
            elif "domain" in column_name or format_name == "domain":
                field_args["validate"].append(validate_by_type("domain"))
            elif format_name == "ipv4":
                field_args["validate"].append(validate_by_type("ipv4"))
            elif format_name == "ipv6":
                field_args["validate"].append(validate_by_type("ipv6"))
            elif format_name == "mac":
                field_args["validate"].append(validate_by_type("mac"))
            elif format_name == "slug":
                field_args["validate"].append(validate_by_type("slug"))
            elif format_name == "uuid" or column.type.python_type == uuid.UUID:
                field_args["validate"].append(validate_by_type("uuid"))
            elif format_name == "card":
                field_args["validate"].append(validate_by_type("card"))
            elif format_name == "country_code":
                field_args["validate"].append(validate_by_type("country_code"))

        return field_args

    def get_url(self, obj, attribute, other_schema):
        """
        Get the URL for a related object.
        Args:
            obj (Any): The object to get the URL from.
            attribute (str): The attribute name to get the URL for.

        Returns:

        """
        related = getattr(obj, attribute)
        if isinstance(related, list):
            return [item.to_url() for item in related]
        elif related:
            return related.to_url()
        else:
            return None

    def get_many_url(self, obj, attribute, other_schema):
        """
        Get the URL for many related object.
        Args:
            obj (Any): The object to get the URL from.
            attribute (str): The attribute name to get the URL for.

        Returns:

        """

        child_end = get_config_or_model_meta("API_ENDPOINT_NAMER", other_schema.Meta.model, default=endpoint_namer)(other_schema.Meta.model)

        return getattr(obj,  child_end.replace("-", "_") + "_to_url")()


    def add_relationship_field(
        self,
        attribute: str,
        original_attribute: str,
        relationship_property: RelationshipProperty,
    ):
        """Automatically add a field for a given relationship in the SQLAlchemy model."""
        max_depth = 1  # Set the maximum depth
        current_depth = self.context.get('current_depth', 0)

        if current_depth >= max_depth:
            # Exceeded maximum depth, include only the primary key reference
            related_model = relationship_property.mapper.class_
            primary_key = inspect(related_model).primary_key[0]
            field_type = type_mapping.get(type(primary_key.type), fields.Integer)

            self.add_to_fields(
                original_attribute, field_type(data_key=attribute)
            )
            return

        else:
            # Include the nested schema and increment the depth
            input_schema, output_schema = get_input_output_from_model_or_make(
                relationship_property.mapper.class_,
                context={'current_depth': current_depth + 1}
            )

            # Access the RelationshipProperty to get 'viewonly' and 'uselist'
            relationship_prop = relationship_property.property
            field_args = {"dump_only": not relationship_prop.viewonly}

            # Determine the serialization type
            dump_type = get_config_or_model_meta("API_SERIALIZATION_TYPE", self.model, default="url")

            if dump_type == "url":
                if relationship_prop.uselist:
                    # Serialize as a list of URLs
                    self.add_to_fields(
                        original_attribute,
                            fields.Function(lambda obj: self.get_many_url(obj, attribute, input_schema), **field_args)
                        ,
                    )
                else:
                    # Serialize as a single URL
                    self.add_to_fields(
                        original_attribute,
                        fields.Function(lambda obj: self.get_url(obj, attribute, input_schema), **field_args),
                    )
            elif dump_type == "json":
                if relationship_prop.uselist:
                    # Always dump as a list of nested schemas
                    self.add_to_fields(
                        original_attribute,
                        fields.List(fields.Nested(output_schema), **field_args),
                    )
                else:
                    # Always dump as a nested schema
                    self.add_to_fields(
                        original_attribute,
                        fields.Nested(output_schema, **field_args),
                    )
            elif dump_type == "hybrid":
                if relationship_prop.uselist:
                    # Serialize as a list of URLs
                    self.add_to_fields(
                        original_attribute,
                            fields.Function(lambda obj: self.get_many_url(obj, attribute, input_schema), **field_args))
                else:
                    # Serialize as a nested schema
                    self.add_to_fields(
                        original_attribute,
                        fields.Nested(output_schema, **field_args),
                    )
            else:
                # Fallback to JSON serialization if an unknown dump_type is provided
                if relationship_prop.uselist:
                    self.add_to_fields(
                        original_attribute,
                        fields.List(fields.Nested(output_schema), **field_args),
                    )
                else:
                    self.add_to_fields(
                        original_attribute,
                        fields.Nested(output_schema, **field_args),
                    )


        self._update_field_metadata(original_attribute)

    def _update_field_metadata(self, attribute: str):
        """Update metadata for the generated field."""
        field_obj = self.fields[attribute]  # Get the Marshmallow field object
        field_meta = field_obj.metadata  # Extract the existing metadata

        # Correctly call get_openapi_meta_data with the field object
        openapi_meta_data = get_openapi_meta_data(field_obj)

        # If the function returns metadata, update the field's metadata
        if openapi_meta_data:
            field_meta.update(openapi_meta_data)

    def dump(self, obj, *args, **kwargs):
        # print("Data before super().dump:", obj)
        if self.fields:
            result = super().dump(obj, *args, **kwargs)
        else:
            #todo This really needs looking at, it shouldn't be happening but when we're in a nested schema it is not
            # generating the fields as expected. It produces an empty dict. This is a dirty fix

            result = self.__class__(context=self.context).dump(obj, *args, **kwargs)
        # print("Data after super().dump:", result)
        return result



import pytest

from demo.model_extension.model import create_app
from demo.model_extension.model.models import Author


@pytest.fixture
def app():
    app = create_app({
        'API_TITLE': 'Automated test',
        'API_VERSION': '0.2.0',
        'API_IGNORE_UNDERSCORE_ATTRIBUTES': True,
        # Other configurations specific to this test
    })
    yield app


@pytest.fixture
def client(app):
    return app.test_client()


def test_examples_data_type_and_desc(client):
    info = Author.date_of_birth.info
    description = info.get("description")
    format = info.get("format")
    example = info.get("example")

    swagger_response = client.get('/swagger.json').json

    assert 1 == 2
